import unittest
import sys
import os

# Required to add peakcan
sDirName = os.path.dirname("")
sys.path.append(sDirName)
file_path = "../"
sDirName = os.path.dirname(file_path)
sys.path.append(sDirName)
import time

from mytoolit.old.network import Network
from mytoolit.old.MyToolItCommands import (
    ActiveState,
    byte_list_to_int,
    MyToolItBlock,
    MyToolItStatData,
    MyToolItSystem,
    NodeState,
    NetworkState,
)
from mytoolit.old.MyToolItNetworkNumbers import MyToolItNetworkNr

sLogLocation = "../../Logs/STU/"
"""
This class supports the manual tests of the Stationary Transceiving Unit (STU)
"""


class TestSthManually(unittest.TestCase):
    def setUp(self):
        print("TestCase: ", self._testMethodName)
        input("Press Any Key to Continue")
        self.fileName = sLogLocation + self._testMethodName + ".txt"
        self.fileNameError = (
            sLogLocation + "Error_" + self._testMethodName + ".txt"
        )
        self.Can = Network(
            sender=MyToolItNetworkNr["SPU1"],
            receiver=MyToolItNetworkNr["STH1"],
        )
        self.Can.logger.info("TestCase: " + str(self._testMethodName))
        self._resetStu()
        self.bError = False
        self.Can.logger.info(
            "STU BlueTooth Address: "
            + hex(self.Can.BlueToothAddress(MyToolItNetworkNr["STU1"]))
        )
        self._statusWords()
        print("Start")
        self.Can.logger.info("Start")

    def tearDown(self):
        if False != self.Can.bError:
            self.bError = True
        self.Can.__exit__()
        if self._test_has_failed():
            if os.path.isfile(self.fileNameError) and os.path.isfile(
                self.fileName
            ):
                os.remove(self.fileNameError)
            if os.path.isfile(self.fileName):
                os.rename(self.fileName, self.fileNameError)

    """
    Checks if any test has failed
    """

    def _test_has_failed(self):
        for _method, error in self._outcome.errors:
            if error:
                return True
        if True == self.bError:
            return True
        return False

    """
    Resets the STU
    """

    def _resetStu(self, retries=5, log=True):
        return self.Can.reset_node("STU1", retries=retries, log=log)

    """
    Get all status words of the STU
    """

    def _statusWords(self):
        psw0 = self.Can.node_status(MyToolItNetworkNr["STU1"])
        self.Can.logger.info("STU Status Word: " + hex(psw0))
        self.Can.logger.info(
            "STU Error Word: {}".format(
                self.Can.error_status(MyToolItNetworkNr["STU1"])
            )
        )

    """
    Test Acknowledgement from STH. Write message and check identifier to be ack (No bError)
    """

    def testManually0001Ack(self):
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 1, 0
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STU1"], [0]
        )
        self.Can.logger.info("Write Message")
        self.Can.WriteFrame(msg)
        self.Can.logger.info("Wait 200ms")
        time.sleep(0.25)
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 0, 0
        )
        msgAckExpected = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["STU1"], MyToolItNetworkNr["SPU1"], [0]
        )
        self.Can.logger.info(
            "Send ID: "
            + hex(msg.ID)
            + "; Expected ID: "
            + hex(msgAckExpected.ID)
            + "; Received ID: "
            + hex(self.Can.getReadMessage(-1).ID)
        )
        expectedData = ActiveState()
        expectedData.asbyte = 0
        expectedData.b.u2NodeState = NodeState["Application"]
        expectedData.b.u3NetworkState = NetworkState["Operating"]
        self.Can.logger.info(
            "Send Data: "
            + hex(0)
            + "; Expected Data: "
            + hex(expectedData.asbyte)
            + "; Received Data: "
            + hex(self.Can.getReadMessage(-1).DATA[0])
        )
        self.assertEqual(
            hex(msgAckExpected.ID), hex(self.Can.getReadMessage(-1).ID)
        )
        self.assertEqual(
            expectedData.asbyte, self.Can.getReadMessage(-1).DATA[0]
        )

    """
    Under Voltage Counter
    """

    def testManually0700PowerOnOffCounter(self):
        PowerOnOff1 = self.Can.statisticalData(
            MyToolItNetworkNr["STU1"], MyToolItStatData["PocPof"]
        )
        PowerOn1 = byte_list_to_int(PowerOnOff1[:4])
        PowerOff1 = byte_list_to_int(PowerOnOff1[4:])
        self.Can.logger.info(
            "Power On Counter since first Power On: " + str(PowerOn1)
        )
        self.Can.logger.info(
            "Power Off Counter since first Power On: " + str(PowerOff1)
        )
        input(
            "Power Off Device, wait 1s, power on again and then press Any Key"
            " to Continue"
        )
        PowerOnOff2 = self.Can.statisticalData(
            MyToolItNetworkNr["STU1"], MyToolItStatData["PocPof"]
        )
        PowerOn2 = byte_list_to_int(PowerOnOff2[:4])
        PowerOff2 = byte_list_to_int(PowerOnOff2[4:])
        self.Can.logger.info(
            "Power On Counter since first Power On: " + str(PowerOn2)
        )
        self.Can.logger.info(
            "Power Off Counter since first Power On: " + str(PowerOff2)
        )
        self.assertEqual(PowerOn1 + 1, PowerOn2)
        self.assertEqual(PowerOff1 + 1, PowerOff2)


if __name__ == "__main__":
    print(sys.version)
    sLogLocation = sys.argv[1]
    sLogFile = sys.argv[2]
    if "/" != sLogLocation[-1]:
        sLogLocation += "/"
    sLogFileLocation = sLogLocation + sLogFile
    sDirName = os.path.dirname(sLogFileLocation)
    sys.path.append(sDirName)

    print("Log Files will be saved at: " + str(sLogFileLocation))
    if not os.path.exists(sDirName):
        os.makedirs(sDirName)
    with open(sLogFileLocation, "w") as f:
        print(f)
        runner = unittest.TextTestRunner(f)
        unittest.main(argv=["first-arg-is-ignored"], testRunner=runner)



from sys import stderr
from time import sleep
from typing import Callable, Iterable, Tuple

from curses import curs_set, error, wrapper  # type: ignore[attr-defined]

from mytoolit.can.sensor import SensorConfig
from mytoolit.old.cli import CommandLineInterface
from mytoolit.old.MyToolItCommands import (
    AdcAcquisitionTime,
    AdcOverSamplingRate,
    AdcReference,
    byte_list_to_int,
    CalibMeassurementActionNr,
    CalibMeassurementTypeNr,
    int_to_mac_address,
    MyToolItStreaming,
)
from mytoolit.old.MyToolItNetworkNumbers import MyToolItNetworkNr
from mytoolit.old.MyToolItSth import fVoltageBattery
from mytoolit.old.network import UnsupportedFeatureException


class Key:
    """Store key constants"""

    CTRL_C = 3
    DELETE = 8
    RETURN = 10

    ZERO = ord("0")
    ONE = ord("1")
    NINE = ord("9")

    A = ord("a")
    C = ord("c")
    F = ord("f")
    N = ord("n")
    P = ord("p")
    Q = ord("q")
    R = ord("r")
    S = ord("s")

    UPPERCASE_O = ord("O")

    ENTER = 459


class UserInterface(CommandLineInterface):
    """ICOc command line & curses interface

    This class can be used to connect to the ICOtronic system and acquire
    measurement data using

    - a command line interface and
    - a menu based curses interface.

    """

    def __init__(self):
        super().__init__()

    def close(self):
        super().close()

    def add_string(self, *arguments) -> None:
        """Add a string to the terminal window

        This is basically a wrapper around the `addstr` method that ignores
        errors (about drawing strings outside the visible area)

        Parameters
        ----------

        arguments:
            A list of argument for the `addstr` method

        """

        try:
            self.stdscr.addstr(*arguments)
        except error:
            # Ignore errors about drawing strings outside terminal window
            pass

    def read_input(
        self,
        allowed_key: Callable[[int], bool],
        allowed_value: Callable[[str], bool],
        default: str = "",
    ) -> Tuple[bool, str]:
        """Read textual input at the current position

        The function will read input until

        - the user enters an allowed value followed by the Enter/Return key, or
        - the user enters `Ctrl` + `C` to stop the input reading.

        Parameters
        ----------

        allowed_key:
            A function that specifies if a given key character is allowed to
            occur in the input or not

        allowed_value:
            A function that specifies if the whole read input is an allowed
            value or not

        default:
            This text will be be used as initial value for the input

        Returns
        -------

        A pair containing:

        - a boolean that specifies if the input is valid according to the
          function `allowed_value`, and
        - the read number.

        """

        y_position, x_position = self.stdscr.getyx()

        text = default
        while True:
            self.add_string(y_position, x_position, text)
            self.stdscr.refresh()
            key = self.stdscr.getch()
            sleep(0.001)  # Sleep a little bit to reduce CPU usage

            if key == Key.CTRL_C:
                break

            if key in {Key.RETURN, Key.ENTER}:
                if allowed_value(text):
                    break
            elif allowed_key(key):
                text += chr(key)
            elif key == Key.DELETE:
                text = text[:-1] if len(text) > 1 else ""
                self.add_string(y_position, x_position + len(text), " ")
                self.stdscr.refresh()

        self.add_string("\n")
        self.stdscr.refresh()
        return (allowed_value(text), text)

    def read_number(
        self,
        default: int = 0,
        allowed: Callable[[int], bool] = lambda value: True,
    ) -> Tuple[bool, int]:
        """Read a number at the current position

        The function will read input until

        - the user enters an allowed number followed by the Enter/Return key,
          or
        - the user enters `Ctrl` + `C` to stop the input reading.

        Parameters
        ----------

        default:
            A number that will be used as initial value

        allowed:
            A function that determines if the current read value is valid or
            not

        Returns
        -------

        A pair containing:

        - a boolean that specifies if the read number is valid according to the
          function `allowed`, and
        - the read number.

        """

        valid, number_text = self.read_input(
            allowed_key=lambda key: ord("0") <= key <= ord("9"),
            allowed_value=lambda value: allowed(int(value)),
            default=str(default),
        )

        return (valid, int(number_text))

    def read_text(
        self,
        default: str = "",
        allowed: Callable[[str], bool] = lambda value: True,
    ) -> Tuple[bool, str]:
        """Read a text at the current position

        The function will read input until

        - the user enters an allowed value followed by the Enter/Return key, or
        - the user enters `Ctrl` + `C` to stop the input reading.

        Parameters
        ----------

        default:
            The initial value for the text

        allowed:
            A function that determines if the current read text is valid or
            not

        Returns
        -------

        A pair containing:

        - a boolean that specifies if the read text is valid according to the
          function `allowed`, and
        - the read number.

        """

        return self.read_input(
            allowed_key=lambda key: ord(" ") <= key <= ord("~"),
            allowed_value=allowed,
            default=default,
        )

    def draw_menu(self, choices: Iterable[str]) -> None:
        """Draw a menu displaying the given choices at the current location

        Parameters
        ----------

        choices:
            A list containing the different choices available

        """

        ruler = "─" * (max(map(len, choices)) + 2)
        max_length_choices = max(map(len, choices))

        self.add_string(f"\n┌{ruler}┐\n")

        for choice in choices:
            fill = max_length_choices - len(choice)
            self.add_string(f"│ {choice}{' '*fill} │\n")

        self.add_string(f"└{ruler}┘")

        self.stdscr.refresh()

    def change_adc_values_window(self):
        def list_keys(dictionary):
            keys = map(str, dictionary.keys())
            return ", ".join(keys)

        def read_value(description, default, allowed):
            self.add_string(description)
            self.stdscr.refresh()

            valid_input, value = (
                self.read_number(default, allowed)
                if isinstance(default, int)
                else self.read_text(default, allowed)
            )

            if not valid_input:
                self.add_string(
                    f"“{value}” is not a valid value; "
                    f"Using default value “{default}” instead\n"
                )
                value = default

            return value

        curs_set(True)
        self.stdscr.clear()

        prescalar = read_value(
            "Prescaler (2–127): ", 2, lambda value: 2 <= value <= 127
        )
        acquistion_time = read_value(
            f"Acquisition Time ({list_keys(AdcAcquisitionTime)}): ",
            8,
            lambda value: value in AdcAcquisitionTime,
        )
        oversampling_rate = read_value(
            f"Oversampling Rate ({list_keys(AdcOverSamplingRate)}): ",
            64,
            lambda value: value in AdcOverSamplingRate,
        )

        adc_reference = read_value(
            f"ADC Reference Voltage (VDD=3V3) ({list_keys(AdcReference)}): ",
            "VDD",
            lambda value: value in AdcReference,
        )

        self.vAdcConfig(prescalar, acquistion_time, oversampling_rate)
        self.vAdcRefVConfig(adc_reference)

    def change_sensors_window(self):
        """Update sensor configuration"""

        def read_channel_value(channel):
            self.add_string(
                "Sensor number (1 – 255) for measurement "
                f"channel {channel} (0 to disable): "
            )

            value = self.read_input(
                default=str(channel),
                allowed_key=lambda key: ord("0") <= key <= ord("9"),
                allowed_value=lambda value: len(value) <= 3
                and int(value) <= 255,
            )[1]

            return int(value)

        def enable_disable_sensors(channel, default):
            self.add_string(
                f"Enable measurement channel {channel} "
                "(1 to enable, 0 to disable): "
            )

            value = self.read_input(
                default=str(default),
                allowed_key=lambda key: ord("0") <= key <= ord("1"),
                allowed_value=lambda value: len(value) == 1,
            )[1]

            return int(value)

        curs_set(True)
        self.stdscr.clear()

        if self.channel_config_supported:
            sensors = [read_channel_value(channel) for channel in range(1, 4)]
            # We use sensor number 1 for disabled sensors
            self.Can.write_sensor_config(*sensors)
        else:
            enabled_sensors = map(
                lambda channel_number: int(bool(channel_number)),
                (self.sensor.first, self.sensor.second, self.sensor.third),
            )
            sensors = [
                enable_disable_sensors(channel, default_value)
                for channel, default_value in enumerate(
                    enabled_sensors, start=1
                )
            ]

        # Enable/disable axes for transmission
        self.set_sensors(*sensors)

    def change_runtime_window(self):
        curs_set(True)
        self.stdscr.clear()

        self.add_string(
            "Run time of data acquisition "
            "(in seconds; 0 for infinite runtime): "
        )
        self.stdscr.refresh()
        runtime = self.read_number()[1]
        self.vRunTime(runtime)

    def sth_window_information(self):
        def read_voltage() -> str:
            index = self.Can.singleValueCollect(
                MyToolItNetworkNr["STH1"],
                MyToolItStreaming["Voltage"],
                1,
                0,
                0,
                log=False,
            )
            iBatteryVoltage = byte_list_to_int(
                self.Can.getReadMessageData(index)[2:4]
            )
            if iBatteryVoltage is not None:
                return f"{fVoltageBattery(iBatteryVoltage):4.2f}"

            return "–"

        def read_temperature() -> float:
            au8TempReturn = self.Can.calibMeasurement(
                MyToolItNetworkNr["STH1"],
                CalibMeassurementActionNr["Measure"],
                CalibMeassurementTypeNr["Temp"],
                1,
                AdcReference["1V25"],
                log=False,
            )
            self.Can.calibMeasurement(
                MyToolItNetworkNr["STH1"],
                CalibMeassurementActionNr["None"],
                CalibMeassurementTypeNr["Temp"],
                1,
                AdcReference["VDD"],
                log=False,
                bReset=True,
            )
            iTemperature = float(byte_list_to_int(au8TempReturn[4:]))
            return iTemperature / 1000

        self.window_header()

        address = int_to_mac_address(int(self.iAddress, 16))
        name = self.sth_name
        device_description = f"STH “{name}” ({address})"
        for value in (
            device_description,
            "\n",
            "—" * len(device_description),
            "\n\n",
        ):
            self.add_string(value)

        hardware_version = self.Can.sProductData(
            "Hardware Version", bLog=False
        )
        software_version = self.Can.sProductData(
            "Firmware Version", bLog=False
        )
        release_name = self.Can.sProductData("Release Name", bLog=False)
        serial_number = self.Can.sProductData("Serial Number", bLog=False)
        product_name = self.Can.sProductData("Product Name", bLog=False)
        serial = f"{serial_number}–{product_name}"
        sensor_range, success = self.read_acceleration_range()
        sensor_range_output = f"± {sensor_range/2} g"
        if not success:
            sensor_range_output += " (EEPROM Read Error)"

        infos = [
            ("Hardware Version", hardware_version),
            ("Firmware Version", software_version),
            ("Firmware Release Name", release_name),
            ("Serial Number", serial),
            ("Sensor Range", f"{sensor_range_output}\n"),
        ]

        voltage = read_voltage()
        temperature = read_temperature()

        infos.extend(
            [
                ("Supply Voltage", f"{voltage} V"),
                ("Chip Temperature", f"{temperature:4.1f} °C\n"),
            ]
        )

        runtime = "∞" if self.iRunTime == 0 else str(self.iRunTime)
        prescaler = self.iPrescaler
        acquistion_time = AdcAcquisitionTime.inverse[self.iAquistionTime]
        oversampling_rate = AdcOverSamplingRate.inverse[self.iOversampling]
        sampling_rate = self.samplingRate
        adc_reference = self.sAdcRef

        sensor_config_local = (
            self.sensor.first,
            self.sensor.second,
            self.sensor.third,
        )

        if self.channel_config_supported:
            sensor_config_device = self.Can.read_sensor_config()
            sensor_config_device.disable_channel(
                *(not sensor_number for sensor_number in sensor_config_local)
            )
        else:
            sensor_config_device = SensorConfig(
                *[
                    number if enabled else 0
                    for number, enabled in enumerate(
                        sensor_config_local, start=1
                    )
                ]
            )
        sensors = str(sensor_config_device)

        infos.extend(
            [
                ("Run Time", f"{runtime} s\n"),
                ("Prescaler", prescaler),
                ("Acquisition Time", acquistion_time),
                ("Oversampling Rate", oversampling_rate),
                ("Sampling Rate", sampling_rate),
                ("Reference Voltage", f"{adc_reference}\n"),
                ("Sensors", sensors),
            ]
        )

        max_description_length = max(
            len(description) for description, _ in infos
        )
        for description, value in infos:
            fill = max_description_length - len(description) + 1
            self.add_string(f"{description}{' '*fill}{value}\n")

    def sth_window_menu(self):
        self.draw_menu(
            [
                "s: Start Data Acquisition",
                "",
                "n: Change STH Name",
                "r: Change Run Time",
                "a: Configure ADC",
                "p: Configure Sensors",
                "O: Set Standby Mode",
                "",
                "q: Disconnect from STH",
            ]
        )

    def sth_window_key_evaluation(self) -> Tuple[bool, bool]:
        """Evaluate key input in the STH window

        Returns
        -------

        A tuple containing boolean values, where

        - the first value specifies if the program should stay in the STH
          window, and
        - the second value specifies if the program execution should continue
          (in the main menu)

        """

        curs_set(False)

        continue_sth_window = True
        continue_program = True

        key = self.stdscr.getch()

        if key == Key.CTRL_C:
            continue_program = False
            continue_sth_window = False
        elif key == Key.A:
            self.change_adc_values_window()
        elif key == Key.P:
            self.change_sensors_window()
        elif key == Key.Q:
            continue_sth_window = False
            self.Can.bBlueToothDisconnect(MyToolItNetworkNr["STU1"])
        elif key == Key.N:
            self.change_sth_name_window()
        elif key == Key.UPPERCASE_O:
            self.stdscr.clear()
            self.add_string("Are you really sure?\n")
            self.add_string("Only charing will leave this state!\n")
            self.add_string("Pressing “y” will trigger standby: ")
            curs_set(True)
            self.stdscr.refresh()

            if self.read_text()[1] == "y":
                self.Can.Standby(MyToolItNetworkNr["STH1"])
                self.Can.bBlueToothDisconnect(MyToolItNetworkNr["STU1"])
                continue_sth_window = False
        elif key == Key.R:
            self.change_runtime_window()
        elif key == Key.S:
            self.stdscr.clear()
            self.add_string("Collecting measurement data…")
            self.stdscr.refresh()
            if not self.KeyBoardInterrupt:
                try:
                    self.vDataAquisition()
                except KeyboardInterrupt:
                    self.KeyBoardInterrupt = True
                    self.__exit__()
                continue_sth_window = False
                continue_program = False

        return (continue_sth_window, continue_program)

    def sth_window(self):
        continue_main = True
        continue_sth = True

        self.update_sensor_config()

        while continue_sth:
            self.sth_window_information()
            self.sth_window_menu()
            continue_sth, continue_main = self.sth_window_key_evaluation()

        return continue_main

    def connect_sth_window(self, number):
        curs_set(True)

        while True:
            devices = self.main_window_information()

            self.add_string("\n")
            y_position = self.stdscr.getyx()[0]
            self.add_string(
                f"Choose STH number (Use “return” to connect): {number}"
            )
            self.stdscr.refresh()
            key = self.stdscr.getch()

            if ord("0") <= key <= ord("9"):
                digit = int(key) - ord("0")
                number = number * 10 + digit

            elif key == Key.DELETE:
                number = int(str(number)[:-1]) if len(str(number)) > 1 else 0

            elif key in {Key.RETURN, Key.ENTER}:
                curs_set(False)

                device_number = number - 1
                device = None
                for dev in devices:
                    if dev["DeviceNumber"] == device_number:
                        device = dev

                if not device:
                    return False

                name = device["Name"]
                self.add_string(
                    y_position, 0, f"Connecting to device “{name}”…{' '*20}"
                )
                self.stdscr.refresh()

                self.vDeviceAddressSet(hex(device["Address"]))
                self.sth_name = name
                self.stdscr.refresh()
                success = self.Can.bBlueToothConnectPollingAddress(
                    MyToolItNetworkNr["STU1"], self.iAddress
                )
                if not success:
                    return False

                # Check if the connected hardware supports channel
                # configuration
                try:
                    self.Can.read_sensor_config()
                    self.channel_config_supported = True
                except UnsupportedFeatureException:
                    self.channel_config_supported = False

                return True

            elif key in {Key.CTRL_C, ord("q")}:
                return False

        return False

    def change_filename_window(self):
        curs_set(True)
        self.stdscr.clear()

        self.add_string("Set base output file name: ")
        self.stdscr.refresh()

        forbidden_characters = set('<>:"/\\|?*')
        input_valid, filename = self.read_text(
            default=str(self.output_filename.stem),
            allowed=lambda filename: 1 <= len(filename) <= 200 >= 1
            and not set(filename).intersection(forbidden_characters),
        )

        curs_set(False)
        if input_valid:
            self.set_output_filename(filename)
        self.add_string(
            "New full name (including time stamp): "
            f"“{self.get_output_filepath()}”"
        )
        self.stdscr.refresh()
        sleep(2)

    def change_sth_name_window(self):
        curs_set(True)
        self.stdscr.clear()
        self.add_string("New STH name (max. 8 characters): ")
        self.stdscr.refresh()
        name_valid, name = self.read_text(
            default=self.sth_name, allowed=lambda text: len(text) <= 8
        )

        if not name_valid:
            return

        self.sth_name = name
        self.Can.vBlueToothNameWrite(MyToolItNetworkNr["STH1"], 0, name)

    def window_header(self):
        self.stdscr.clear()
        self.add_string(f"{' '*16}ICOc\n\n")

    def main_window_information(self):
        self.window_header()

        # Removing a disconnected sensor device from the list of displayed
        # sensor devices takes quite some time (about 40 seconds). The first
        # reason for that is probably the function `tDeviceList` below, which
        # takes about 20 seconds before it removes a disconnected sensor from
        # the returned list. It takes about another 20 seconds before the list
        # of displayed sensors is updated afterwords.
        #
        # The reason behind this bug is unclear, especially since updating the
        # list with new values usually takes less than a second. Adding a new
        # or recently disconnected sensor device to the list also seems to work
        # without any perceptible delay.

        devices = self.Can.tDeviceList(MyToolItNetworkNr["STU1"], bLog=False)
        sleep(0.02)  # Sleep a little bit to reduce CPU usage

        header = f"{' '*7}Name      Address            RSSI{' '*7}"
        ruler = "—" * len(header)
        for line in (header, ruler):
            self.add_string(f"{line}\n")

        for device in devices:
            number = device["DeviceNumber"] + 1
            address = int_to_mac_address(device["Address"])
            name = device["Name"]
            rssi = device["RSSI"]
            self.add_string(f"{number:5}: {name:8}  {address}  {rssi} dBm\n")

        return devices

    def main_window_menu(self):
        self.draw_menu(
            [
                "1-9: Connect to STH",
                "",
                "  f: Change Output File Name",
                "  n: Change STH Name",
                "",
                "  q: Quit ICOc",
            ]
        )

    def main_window_key_evaluation(self) -> bool:
        """Evaluate key input in the main window

        Returns
        -------

        - True, if the execution of the main menu should continue
        - False, if the user interface should be closed

        """

        key = self.stdscr.getch()

        if key in {Key.Q, Key.CTRL_C}:
            return False

        if Key.ONE <= key <= Key.NINE:
            return (
                self.sth_window()
                if self.connect_sth_window(int(key - Key.ZERO))
                else True
            )

        if key == Key.F:
            self.change_filename_window()
            return True

        if key == Key.N:
            if self.connect_sth_window(0):
                self.change_sth_name_window()
                self.Can.bBlueToothDisconnect(MyToolItNetworkNr["STU1"])
            else:
                self.add_string("Device was not available\n")
                self.stdscr.refresh()

        return True

    def main_window(self):
        curs_set(False)

        self.main_window_information()
        self.main_window_menu()

        # Refresh display to remove disconnected sensor devices
        self.stdscr.refresh()

        return self.main_window_key_evaluation()

    def user_interface(self, stdscr):
        self.stdscr = stdscr

        # TODO: Do not refresh the whole display constantly
        # Possible Solution:
        # - Spawn two threads
        # - One of them waits for input (blocking)
        # - Other thread refreshes list of devices
        self.stdscr.nodelay(1)
        self.stdscr.clear()

        try:
            while self.main_window():
                pass
        except KeyboardInterrupt:
            self.KeyBoardInterrupt = True

    def run(self):
        self._vRunConsoleStartup()
        self.reset()
        if self.connect:
            self.vRunConsoleAutoConnect()
        else:
            wrapper(self.user_interface)
        self.close()


def main():
    try:
        UserInterface().run()
    except Exception as error:
        print(f"Error\n—————\n☹️ {error}\n", file=stderr)
        print("Stack Trace\n———————————", file=stderr)
        raise error


if __name__ == "__main__":
    main()



# -- Imports ------------------------------------------------------------------

from argparse import ArgumentParser

from mytoolit.cmdline import base64_mac_address
from mytoolit.utility import convert_base64_mac

# -- Main ---------------------------------------------------------------------


def main():
    parser = ArgumentParser(
        description="Convert the Base64 name of an STH to a MAC address"
    )
    parser.add_argument(
        "name", help="name of the STH e.g. CGvXAd6B", type=base64_mac_address
    )
    name = parser.parse_args().name
    mac = convert_base64_mac(name)
    print(mac)


if __name__ == "__main__":
    main



# -- Imports ------------------------------------------------------------------

from time import sleep

from mytoolit.config import settings
from mytoolit.test.production import TestNode

# -- Classes ------------------------------------------------------------------


class TestSensorNode(TestNode):
    """This class contains support code for sensor node (SMH & STH)

    You are not supposed to use this class directly, but instead use it as
    superclass for your test class. For more information, please take a look
    at the documentation of `TestNode`.

    """

    @classmethod
    def setUpClass(cls):
        """Set up data for whole test"""

        super().setUpClass()

        cls.add_attribute("Serial Number", "{cls.serial_number}", pdf=True)
        cls.add_attribute(
            "Ratio Noise Maximum", "{cls.ratio_noise_max:.3f} dB"
        )
        cls.add_attribute("Sleep Time 1", "{cls.sleep_time_1} ms", pdf=False)
        cls.add_attribute(
            "Advertisement Time 1", "{cls.advertisement_time_1} ms", pdf=False
        )
        cls.add_attribute("Sleep Time 2", "{cls.sleep_time_2} ms", pdf=False)
        cls.add_attribute(
            "Advertisement Time 2", "{cls.advertisement_time_2} ms", pdf=False
        )

    def _connect_device(self, name: str) -> None:
        """Create a connection to the device with the specified name

        Parameters
        ----------

        name:
            The (Bluetooth advertisement) name of the sensor device

        """

        super()._connect()  # Connect to STU
        self.loop.run_until_complete(self.can.connect_sensor_device(name))

    def _read_data(self):
        """Read data from connected sensor device"""

        async def read_data():
            """Read data using the new network class"""

            node = "STH 1"
            cls.bluetooth_mac = await self.can.get_mac_address(node)
            cls.bluetooth_rssi = await self.can.get_rssi(node)
            cls.firmware_version = await self.can.get_firmware_version(node)

        cls = type(self)
        self.loop.run_until_complete(read_data())

    def _test_connection_device(self):
        """Check connection to sensor device

        This tests sends a command from the STU (with the identifier of SPU1)
        to the STH/Smh and checks if the acknowledgment message from the STH
        contains the same data as the sent message (, except for switched
        sender/receiver and flipped acknowledgment bit).
        """

        # The sensor devices need a little more time to switch from the
        # “Startup” to the “Operating” state
        sleep(1)

        self.loop.run_until_complete(self._test_connection("STH 1"))

    async def _test_eeprom_sleep_advertisement_times(self):
        """Test if reading and writing of sleep/advertisement times works"""

        async def read_write_time(
            read_function, write_function, variable, description, milliseconds
        ):
            await write_function(milliseconds)
            milliseconds_read = round(await read_function())
            setattr(type(self), variable, milliseconds_read)
            self.assertEqual(
                milliseconds_read,
                milliseconds,
                (
                    f"{description} {milliseconds_read} ms does not match "
                    f" written value of {milliseconds} ms"
                ),
            )

        await read_write_time(
            read_function=self.can.read_eeprom_sleep_time_1,
            write_function=self.can.write_eeprom_sleep_time_1,
            variable="sleep_time_1",
            description="Sleep Time 1",
            milliseconds=settings.sensory_device.bluetooth.sleep_time_1,
        )

        await read_write_time(
            read_function=self.can.read_eeprom_advertisement_time_1,
            write_function=self.can.write_eeprom_advertisement_time_1,
            variable="advertisement_time_1",
            description="Advertisement Time 1",
            milliseconds=(
                settings.sensory_device.bluetooth.advertisement_time_1
            ),
        )

        await read_write_time(
            read_function=self.can.read_eeprom_sleep_time_2,
            write_function=self.can.write_eeprom_sleep_time_2,
            variable="sleep_time_2",
            description="Sleep Time 2",
            milliseconds=settings.sensory_device.bluetooth.sleep_time_2,
        )

        await read_write_time(
            read_function=self.can.read_eeprom_advertisement_time_2,
            write_function=self.can.write_eeprom_advertisement_time_2,
            variable="advertisement_time_2",
            description="Advertisement Time 2",
            milliseconds=(
                settings.sensory_device.bluetooth.advertisement_time_2
            ),
        )



# -- Imports ------------------------------------------------------------------

from unittest import main as unittest_main

from mytoolit.can.node import Node
from mytoolit.config import settings
from mytoolit.test.production import TestNode
from mytoolit.report import Report
from mytoolit.test.unit import ExtendedTestRunner

# -- Class --------------------------------------------------------------------


class TestSTU(TestNode):
    """This class contains tests for the Stationary Transceiver Unit (STU)"""

    @classmethod
    def setUpClass(cls):
        """Set up data for whole test"""

        super().setUpClass()
        cls.add_attribute("Serial Number", "{cls.serial_number}", pdf=False)
        cls.report = Report(node="STU")

    def _read_data(self):
        """Read data from connected STU"""

        async def read_data():
            """Read data using the new network class"""

            node = "STU 1"
            cls.bluetooth_mac = await self.can.get_mac_address(node)
            cls.firmware_version = await self.can.get_firmware_version(node)
            cls.release_name = await self.can.get_firmware_release_name(node)

        cls = type(self)
        self.loop.run_until_complete(read_data())

    def test__firmware_flash_disconnected(self):
        """Upload bootloader and application into STU

        Please note the additional underscore in the method name that makes
        sure this test case is executed before all other test cases.

        The text `disconnected` in the method name make sure that the test
        framework does not initialize a connection.

        """

        self._test_firmware_flash(
            node="STU",
            flash_location=settings.stu.firmware.location.flash,
            programmmer_serial_number=(
                settings.stu.programming_board.serial_number
            ),
            chip="BGM111A256V2",
        )

    def test_connection(self):
        """Check connection to STU"""

        self.loop.run_until_complete(self._test_connection("STU 1"))

    def test_eeprom(self):
        """Test if reading and writing the EEPROM works"""

        async def test_eeprom():
            """Test the EERPOM of the STU"""

            cls = type(self)
            receiver = "STU 1"

            # ========
            # = Name =
            # ========

            read_name = await self.can.read_eeprom_name()
            expected_name = "Valerie"
            self.assertEqual(
                read_name,
                expected_name,
                f"Read name “{read_name}” does not match expected name "
                + f"“{expected_name}”",
            )

            cls.name = read_name

            # ================
            # = Product Data =
            # ================

            await self._test_eeprom_product_data(Node(receiver), settings.stu)

            # ==============
            # = Statistics =
            # ==============

            await self._test_eeprom_statistics(
                Node(receiver),
                settings.stu.production_date,
                settings.stu.batch_number,
            )

            # =================
            # = EEPROM Status =
            # =================

            await self._test_eeprom_status(Node(receiver))

        self.loop.run_until_complete(test_eeprom())


# -- Main ---------------------------------------------------------------------


def main():
    unittest_main(
        testRunner=ExtendedTestRunner, module="mytoolit.test.production.stu"
    )


if __name__ == "__main__":
    main()



# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from asyncio import Queue
from numbers import Real
from typing import (
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from can import Listener, Message
from pint import Quantity

from mytoolit.can.identifier import Identifier

# -- Classes ------------------------------------------------------------------


class AsyncStreamBuffer(Listener):
    """Buffer for streaming data"""

    def __init__(self, first: bool, second: bool, third: bool) -> None:
        """Initialize object using the given arguments

        Parameters
        ----------

        first:
            Specifies if the data of the first measurement channel will be
            collected or not

        second:
            Specifies if the data of the second measurement channel will be
            collected or not

        third:
            Specifies if the data of the third measurement channel will be
            collected or not

        """

        # Expected identifier of received streaming messages
        self.identifier = Identifier(
            block="Streaming",
            block_command="Data",
            sender="STH 1",
            receiver="SPU 1",
            request=False,
        )
        self.first = first
        self.second = second
        self.third = third
        self.queue: Queue[StreamingData] = Queue()

    def __aiter__(self) -> AsyncIterator[StreamingData]:
        """Retrieve iterator for collected data

        Returns
        -------

        An iterator over the received streaming data

        """

        return self

    def __anext__(self) -> Awaitable[StreamingData]:
        """Retrieve next stream data object in collected data

        Returns
        -------

        Retrieved streaming data

        """

        return self.queue.get()

    def on_message_received(self, message: Message) -> None:
        """Handle received messages

        Parameters
        ----------

        message:
            The received CAN message

        """

        if message.arbitration_id != self.identifier.value:
            return

        data = message.data
        timestamp = message.timestamp
        raw_values = [
            TimestampedValue(
                value=int.from_bytes(word, byteorder="little"),
                timestamp=timestamp,
                counter=data[1],
            )
            for word in (data[2:4], data[4:6], data[6:8])
        ]

        streaming_data = StreamingData()
        first = self.first
        second = self.second
        third = self.third

        if first and second and third:
            streaming_data.first.append(raw_values[0])
            streaming_data.second.append(raw_values[1])
            streaming_data.third.append(raw_values[2])
        elif first and second:
            streaming_data.first.append(raw_values[0])
            streaming_data.second.append(raw_values[1])
        elif first and third:
            streaming_data.first.append(raw_values[0])
            streaming_data.third.append(raw_values[1])
        elif second and third:
            streaming_data.second.append(raw_values[0])
            streaming_data.third.append(raw_values[1])
        elif first:
            streaming_data.first.extend(raw_values)
        elif second:
            streaming_data.second.extend(raw_values)
        else:
            streaming_data.third.extend(raw_values)

        self.queue.put_nowait(streaming_data)


class StreamingFormat:
    """Support for specifying the data streaming format

    See also: https://mytoolit.github.io/Documentation/#block-streaming
    """

    # Possible number of data sets
    data_set = [0, 1, 3, 6, 10, 15, 20, 30]

    def __init__(
        self,
        *value,
        streaming: Optional[bool] = None,
        width: Optional[int] = 2,
        first: Optional[bool] = None,
        second: Optional[bool] = None,
        third: Optional[bool] = None,
        sets: Optional[int] = None,
        value_explanations: Tuple[str, str, str] = (
            "Value 1",
            "Value 2",
            "Value 3",
        ),
    ) -> None:
        """Initialize the streaming format using the given arguments

        Positional Parameters
        ---------------------

        value:
            The value of the streaming format byte

        Keyword Parameters
        ------------------

        streaming:
            Specifies if this is a request for a stream of data bytes;
            If this value is not set or set to `False`, then the request is
            only for a single value (or set of values).

        width:
            Specifies the width of a single value (either 2 or 3 bytes)

        first:
            Specifies if the first data value should be transmitted or not

        second:
            Specifies if the second data value should be transmitted or not

        third:
            Specifies if the third data value should be transmitted or not

        sets:
            Specifies the number of data sets that should be transmitted

            The value 0 stops the stream. Other possible values for the number
            of sets are 1, 3, 6, 10, 15, 20 and 30.

        value_explanations:
            Three strings used to describe the first, second and third data
            value

        """

        def set_part(start, width, number):
            """Store bit pattern number at bit start of the identifier"""

            streaming_ones = 0xFF
            mask = (1 << width) - 1

            # Set all bits for targeted part to 0
            self.value &= (mask << start) ^ streaming_ones
            # Make sure we use the correct number of bits for number
            number = number & mask
            # Set bits to given value
            self.value |= number << start

        self.value_explanations = value_explanations

        if len(value) > 1:
            raise ValueError("More than one positional argument")

        self.value = value[0] if value else 0

        # =============
        # = Streaming =
        # =============

        if streaming:
            set_part(7, 1, int(streaming))

        # =========
        # = Width =
        # =========

        if width is not None:
            if not (isinstance(width, int) and 2 <= width <= 3):
                raise ValueError(f"Unsupported width value: {width}")

            set_part(6, 1, 1 if width == 3 else 0)

        # =================
        # = Active Values =
        # =================

        for shift, part in enumerate([third, second, first]):
            if part is not None:
                set_part(3 + shift, 1, part)

        # =============
        # = Data Sets =
        # =============

        if sets is not None:
            cls = type(self)

            if sets not in cls.data_set:
                raise ValueError(f"Unsupported number of data sets: {sets}")

            set_part(0, 3, cls.data_set.index(sets))

    def __repr__(self) -> str:
        """Retrieve the textual representation of the streaming format

        Returns
        -------

        A string that describes the streaming format

        Examples
        --------

        >>> StreamingFormat(width=3, first=True, sets=15)
        Single Request, 3 Bytes, 15 Data Sets, Read Value 1

        >>> StreamingFormat(0b001, streaming=True)
        Streaming, 2 Bytes, 1 Data Set

        >>> StreamingFormat(0b110111)
        Single Request, 2 Bytes, 30 Data Sets, Read Value 1, Read Value 2

        """

        streaming = self.value >> 7

        data_sets = self.data_sets()
        data_set_explanation = (
            "Stop Stream"
            if data_sets == 0
            else "{} Data Set{}".format(
                data_sets, "" if data_sets == 1 else "s"
            )
        )

        parts = [
            "Streaming" if streaming else "Single Request",
            f"{self.data_bytes()} Bytes",
            f"{data_set_explanation}",
        ]

        value_selection = (self.value >> 3) & 0b111

        first = value_selection >> 2
        second = value_selection >> 1 & 1
        third = value_selection & 1

        selected_values = [
            f"Read {value_explanation}"
            for selected, value_explanation in zip(
                (first, second, third), self.value_explanations
            )
            if selected
        ]

        value_explanation = (
            ", ".join(selected_values) if selected_values else ""
        )
        if value_explanation:
            parts.append(value_explanation)

        return ", ".join(parts)

    def data_sets(self) -> int:
        """Get the number of data sets of the streaming format

        Returns
        -------

        The number of data sets

        Examples
        --------

        >>> StreamingFormat(width=3, first=True, sets=15).data_sets()
        15

        >>> StreamingFormat(first=True, second=False, sets=3).data_sets()
        3

        """

        data_set_bits = self.value & 0b111
        cls = type(self)

        return cls.data_set[data_set_bits]

    def data_bytes(self) -> int:
        """Get the number of data bytes used for a single value

        Returns
        -------

        The number of data bytes that represent a single streaming value

        Examples
        --------

        >>> StreamingFormat(width=3, first=True, sets=15).data_bytes()
        3

        >>> StreamingFormat(first=True, second=False, width=2).data_bytes()
        2

        """

        return 3 if (self.value >> 6) & 1 else 2


class StreamingFormatVoltage(StreamingFormat):
    """Support for specifying the streaming format of voltage data"""

    def __init__(self, *arguments, **keyword_arguments) -> None:
        """Initialize the voltage streaming format using the given arguments

        value:
            The value of the streaming format byte

        single:
            Specifies if the request was for a single value or not

        width:
            Specifies the width of a single value (either 2 or 3 bytes)

        first:
            Specifies if the first voltage value should be transmitted or not

        second:
            Specifies if the second voltage value should be transmitted or not

        third:
            Specifies if the third voltage value should be transmitted or not

        sets:
            Specifies the number of data sets that should be transmitted

            The value 0 stops the stream. Other possible values for the number
            of sets are 1, 3, 6, 10, 15, 20 and 30.

        """

        super().__init__(
            *arguments,
            **keyword_arguments,
            value_explanations=("Voltage 1", "Voltage 2", "Voltage 3"),
        )

    def __repr__(self) -> str:
        """Retrieve the textual representation of the voltage streaming format

        Returns
        -------

        A string that describes the voltage streaming format

        Examples
        --------

        >>> StreamingFormatVoltage(streaming=True, width=2, second=True,
        ...                        sets=10)
        Streaming, 2 Bytes, 10 Data Sets, Read Voltage 2

        >>> StreamingFormatVoltage(streaming=False, width=3, first=False,
        ...                        second=True, third=True, sets=3)
        Single Request, 3 Bytes, 3 Data Sets, Read Voltage 2, Read Voltage 3

        """

        return super().__repr__()


class TimestampedValue:
    """Store a single (streaming) value and its timestamp"""

    def __init__(
        self,
        timestamp: float,
        value: Union[float, Quantity],
        counter: int,
    ) -> None:
        """Initialize the timestamped value using the given arguments

        Parameters
        ----------

        timestamp:
            The time when the data value was acquired in seconds since the
            epoch

        value:
            The data value either as
            - raw value or
            - value including a given unit and optional quantity

        counter:
            The message counter of the data

        """

        self.timestamp = timestamp
        self.value = value
        self.counter = counter

    def __repr__(self) -> str:
        """Retrieve the textual representation of the value

        Returns
        -------

        A string that describes the timestamped value

        Examples
        --------

        >>> from mytoolit.measurement import celsius, g0

        >>> TimestampedValue(timestamp=123, value=444, counter=1)
        444@123 (1)
        >>> TimestampedValue(timestamp=20, value=10, counter=10)
        10@20 (10)

        >>> TimestampedValue(timestamp=5, value=g0(10), counter=5)
        10 g_0@5 (5)
        >>> TimestampedValue(timestamp=5, value=celsius(10), counter=20)
        10 °C@5 (20)

        """

        value = (
            f"{self.value:~}"
            if isinstance(self.value, Quantity)
            else str(self.value)
        )

        return f"{value}@{self.timestamp} ({self.counter})"

    def default(self) -> Dict[str, Union[float, int, str]]:
        """Serialize the timestamped value

        Returns
        -------

        An object that can be serialized (into JSON)

        Examples
        --------

        >>> from json import dumps
        >>> from mytoolit.measurement import celsius, g0


        >>> TimestampedValue(timestamp=123, value=11, counter=1).default()
        {'timestamp': 123, 'value': 11, 'counter': 1}

        >>> TimestampedValue(timestamp=123, value=g0(2), counter=1
        ...                 ).default() # doctest:+NORMALIZE_WHITESPACE
        {'timestamp': 123, 'value': 2, 'unit': 'standard_gravity',
         'counter': 1}

        >>> timestamped = TimestampedValue(timestamp=5, value=celsius(10),
        ...                                counter=20)
        >>> dumps(timestamped.default()) # doctest:+NORMALIZE_WHITESPACE
        '{"timestamp": 5, "value": 10, "unit": "degree_Celsius",
          "counter": 20}'

        """

        serialized: Dict[str, Union[float, str]] = {
            "timestamp": self.timestamp
        }
        value = self.value
        if isinstance(value, Quantity):
            serialized["value"] = value.magnitude
            serialized["unit"] = f"{value.units}"
        else:
            serialized["value"] = value
        serialized["counter"] = self.counter

        return serialized


class NotHomogeneousException(Exception):
    """Indicates that a channel of streaming data contains heterogenous data"""


class StreamingData:
    """Auxiliary class to store streaming data"""

    def __init__(
        self,
        first: Optional[List[TimestampedValue]] = None,
        second: Optional[List[TimestampedValue]] = None,
        third: Optional[List[TimestampedValue]] = None,
    ) -> None:
        """Initialize the streaming data using the given arguments

        Parameters
        ----------

        first:
            The data points for the first measurement channel

        second:
            The data points for the second measurement channel

        third:
            The data points for the third measurement channel

        """

        self.first = [] if first is None else first
        self.second = [] if second is None else second
        self.third = [] if third is None else third

    def __iter__(self):
        """Retrieve an iterator for the different measurement channels

        Returns
        -------

        An iterator fir the measurement channels of the streaming data

        Examples
        --------

        >>> value1 = TimestampedValue(timestamp=1, value=1, counter=1)
        >>> value2 = TimestampedValue(timestamp=2, value=2, counter=2)
        >>> value3 = TimestampedValue(timestamp=3, value=3, counter=3)
        >>> data = StreamingData([value1], [], [value2, value3])
        >>> collected = []
        >>> for channel in data:
        ...     collected.extend(channel)
        >>> collected
        [1@1 (1), 2@2 (2), 3@3 (3)]

        """

        return iter([self.first, self.second, self.third])

    def __len__(self) -> int:
        """Retrieve the (combined) length of the streaming data

        Returns
        -------

        The amount of stored streaming values

        """

        return len(self.first) + len(self.second) + len(self.third)

    def __repr__(self) -> str:
        """Retrieve the textual representation of streaming data

        Returns
        -------

        A string that describes the ADC streaming data

        Examples
        --------

        >>> value1 = TimestampedValue(timestamp=1, value=1, counter=1)
        >>> value2 = TimestampedValue(timestamp=2, value=2, counter=2)
        >>> value3 = TimestampedValue(timestamp=3, value=3, counter=3)
        >>> StreamingData([], [], [value1, value2, value3])
        1: []
        2: []
        3: [1@1 (1), 2@2 (2), 3@3 (3)]

        >>> value1 = TimestampedValue(timestamp=1, value=1, counter=10)
        >>> value2 = TimestampedValue(timestamp=2, value=2, counter=11)
        >>> value3 = TimestampedValue(timestamp=3, value=3, counter=12)
        >>> StreamingData([], [], [value1, value2, value3])
        1: []
        2: []
        3: [1@1 (10), 2@2 (11), 3@3 (12)]

        """

        representation = []
        for channel, data in enumerate(
            (self.first, self.second, self.third), start=1
        ):
            representation.append(f"{channel}: {data}")

        return "\n".join(representation)

    def is_homogeneous(self) -> bool:
        """Check if the streaming data for every channel is homogeneous

        The data for a channel is considered homogeneous, if all values have
        the same data type (either `float` or `Quantity`)

        Returns
        -------

        True if the data is homogeneous, false otherwise

        Examples
        --------

        >>> from mytoolit.measurement import celsius

        >>> value1 = TimestampedValue(timestamp=1, value=1, counter=10)
        >>> value2 = TimestampedValue(timestamp=2, value=2, counter=11)
        >>> value3 = TimestampedValue(timestamp=3, value=celsius(3),
        ...                           counter=12)
        >>> value4 = TimestampedValue(timestamp=3, value=celsius(-2),
        ...                           counter=13)

        All values have same data type

        >>> StreamingData([value1], [],
        ...               [value2, value1]).is_homogeneous()
        True

        Values for one channel have different data type

        >>> StreamingData([value1, value3], [], []).is_homogeneous()
        False

        Values of each channel have same data type

        >>> StreamingData([value1, value2], [],
        ...               [value3, value4]).is_homogeneous()
        True

        """

        for channel in self:
            if len(channel) <= 0:
                continue

            value_type_first = type(channel[0].value)
            for timestamped in channel[1:]:
                if type(timestamped.value) != value_type_first:
                    return False

        return True

    def default(
        self, compact: bool = False
    ) -> Union[
        Dict[str, List[Dict[str, Union[float, int, str]]]],
        Dict[str, Dict[str, Union[List[float], List[int], str]]],
    ]:
        """Serialize the streaming data

        Converting streaming data in “compact” form only works for homogenous
        data. If the data of one channel is heterogenous and compact is set to
        true, then this method will throw an `NotHomogeneousException`
        exception.

        Arguments
        ---------

        compact:
            Use a “compact” representation of the streaming data (assumes
            homogenous data)

        Returns
        -------

        An object that can be serialized (into JSON)

        Examples
        --------

        >>> value1 = TimestampedValue(timestamp=1, value=1, counter=10)
        >>> value2 = TimestampedValue(timestamp=2, value=2, counter=11)
        >>> value3 = TimestampedValue(timestamp=3, value=3, counter=12)
        >>> streaming_data = StreamingData([], [], [value1, value2, value3])
        >>> streaming_data.default() # doctest:+NORMALIZE_WHITESPACE
        {'third': [{'timestamp': 1, 'value': 1, 'counter': 10},
                   {'timestamp': 2, 'value': 2, 'counter': 11},
                   {'timestamp': 3, 'value': 3, 'counter': 12}]}
        >>> streaming_data.default(
        ...     compact=True) # doctest:+NORMALIZE_WHITESPACE
        {'third': {'timestamps': [1, 2, 3],
                   'values': [1, 2, 3],
                   'counters': [10, 11, 12]}}

        >>> from mytoolit.measurement import celsius

        >>> value1 = TimestampedValue(timestamp=45, value=celsius(10),
        ...                           counter=10)
        >>> value2 = TimestampedValue(timestamp=67, value=celsius(20),
        ...                           counter=11)
        >>> value3 = TimestampedValue(timestamp=89, value=celsius(30),
        ...                           counter=12)
        >>> streaming_data = StreamingData([value1], [value2, value3], [])
        >>> streaming_data.default(
        ...     compact=True) # doctest:+NORMALIZE_WHITESPACE
        {'first': {'timestamps': [45],
                   'values': [10],
                   'unit': 'degree_Celsius',
                   'counters': [10]},
         'second': {'timestamps': [67, 89],
                    'values': [20, 30],
                    'unit': 'degree_Celsius',
                    'counters': [11, 12]}}

        """

        if compact:
            if not self.is_homogeneous():
                raise NotHomogeneousException(
                    "Unable to serialize data in compact form"
                )

            serializable_compact: Dict[
                str, Dict[str, Union[List[float], List[int], str]]
            ] = {}
            for channel, key in zip(
                (self.first, self.second, self.third),
                ("first", "second", "third"),
            ):
                if len(channel) <= 0:
                    continue

                serializable_compact[key] = {}
                serialized_channel = serializable_compact[key]
                first = channel[0]
                serialized_channel["timestamps"] = [
                    timestamped.timestamp for timestamped in channel
                ]

                values: List[float] = []
                if isinstance(first.value, Quantity):
                    for timestamped in channel:
                        assert isinstance(timestamped.value, Quantity)
                        values.append(timestamped.value.magnitude)
                    serialized_channel["values"] = values
                    serialized_channel["unit"] = f"{first.value.units}"
                else:
                    for timestamped in channel:
                        assert isinstance(timestamped.value, Real)
                        values.append(timestamped.value)
                    serialized_channel["values"] = values

                serialized_channel["counters"] = [
                    timestamped.counter for timestamped in channel
                ]
            return serializable_compact

        # Non “compact” form
        serializable = {}
        for channel, key in zip(
            (self.first, self.second, self.third), ("first", "second", "third")
        ):
            if len(channel) > 0:
                serializable[key] = [
                    timestamped.default() for timestamped in channel
                ]
        return serializable

    def extend(self, data: StreamingData) -> None:
        """Add additional streaming

        Parameters
        ----------

        data:
            The streaming data that should be added to this streaming data
            object

        Examples
        --------

        >>> value11 = TimestampedValue(timestamp=11, value=11, counter=1)
        >>> value12 = TimestampedValue(timestamp=12, value=12, counter=2)
        >>> value13 = TimestampedValue(timestamp=13, value=13, counter=3)
        >>> value14 = TimestampedValue(timestamp=14, value=14, counter=4)
        >>> value21 = TimestampedValue(timestamp=21, value=21, counter=5)
        >>> value22 = TimestampedValue(timestamp=22, value=22, counter=6)
        >>> value31 = TimestampedValue(timestamp=31, value=31, counter=7)
        >>> value32 = TimestampedValue(timestamp=32, value=32, counter=8)
        >>> value33 = TimestampedValue(timestamp=33, value=33, counter=9)
        >>> value34 = TimestampedValue(timestamp=34, value=34, counter=10)

        >>> data = StreamingData([value11, value12], [], [value31, value32])
        >>> other = StreamingData([value13, value14], [value21, value22],
        ...                       [value33, value34])
        >>> data.extend(other)
        >>> data
        1: [11@11 (1), 12@12 (2), 13@13 (3), 14@14 (4)]
        2: [21@21 (5), 22@22 (6)]
        3: [31@31 (7), 32@32 (8), 33@33 (9), 34@34 (10)]
        >>> other
        1: [13@13 (3), 14@14 (4)]
        2: [21@21 (5), 22@22 (6)]
        3: [33@33 (9), 34@34 (10)]

        """

        self.first.extend(data.first)
        self.second.extend(data.second)
        self.third.extend(data.third)

    def empty(self) -> bool:
        """Check if the object contains any streaming data

        Returns
        -------

        `True` if the current streaming data object stores any data, `False`
        otherwise

        Examples
        --------

        >>> StreamingData().empty()
        True
        >>> StreamingData([], [], []).empty()
        True
        >>> StreamingData([], [],
        ...     [TimestampedValue(timestamp=1, value=1, counter=1)]).empty()
        False

        """

        return not (self.first or self.second or self.third)

    def apply(
        self,
        function: Callable[[Union[float, Quantity]], Union[float, Quantity]],
        first: bool = True,
        second: bool = True,
        third: bool = True,
    ) -> None:
        """Apply a certain function to the streaming data

        Parameters
        ----------

        function:
            The function that should be applied to the streaming data

        first:
            Specifies if the function should be applied to the first
            measurement channel or not

        second:
            Specifies if the function should be applied to the second
            measurement channel or not

        third:
            Specifies if the function should be applied to the third
            measurement channel or not

        Examples
        --------

        >>> value11 = TimestampedValue(timestamp=11, value=11, counter=5)
        >>> value12 = TimestampedValue(timestamp=12, value=12, counter=6)
        >>> value13 = TimestampedValue(timestamp=13, value=13, counter=7)
        >>> value14 = TimestampedValue(timestamp=14, value=14, counter=8)

        >>> data = StreamingData([value11, value12], [value13], [value14])
        >>> data
        1: [11@11 (5), 12@12 (6)]
        2: [13@13 (7)]
        3: [14@14 (8)]
        >>> data.apply(lambda value: value + 10)
        >>> data
        1: [21@11 (5), 22@12 (6)]
        2: [23@13 (7)]
        3: [24@14 (8)]

        >>> value11 = TimestampedValue(timestamp=11, value=11, counter=11)
        >>> value12 = TimestampedValue(timestamp=12, value=12, counter=12)
        >>> data = StreamingData([value11, value12])
        >>> data
        1: [11@11 (11), 12@12 (12)]
        2: []
        3: []
        >>> data.apply(lambda value: value*2)
        >>> data
        1: [22@11 (11), 24@12 (12)]
        2: []
        3: []

        """

        def map_list(function, channel):
            return [
                TimestampedValue(
                    timestamped.timestamp,
                    function(timestamped.value),
                    timestamped.counter,
                )
                for timestamped in channel
            ]

        if first:
            self.first = map_list(function, self.first)
        if second:
            self.second = map_list(function, self.second)
        if third:
            self.third = map_list(function, self.third)


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from datetime import datetime
from pathlib import Path
from types import TracebackType
from typing import Dict, Iterable, Optional, Type, Union

from pint import Quantity
from tables import (
    File,
    Filters,
    Float32Col,
    IsDescription,
    MetaAtom,
    MetaIsDescription,
    Node,
    NoSuchNodeError,
    open_file,
    UInt8Col,
    UInt64Col,
)
from tables.exceptions import HDF5ExtError

from mytoolit.can.streaming import StreamingData

# -- Functions ----------------------------------------------------------------


def create_acceleration_description(
    attributes: Dict[str, MetaAtom]
) -> MetaIsDescription:
    """Create a new `IsDescription` class to store acceleration data

    Parameters
    ----------

    attributes:
        A dictionary containing additional columns to store specific
        acceleration data. The key specifies the name of the attribute, while
        the value specifies the type.

    Examples
    --------

    >>> description_x_acceleration = create_acceleration_description(
    ...     dict(x=Float32Col()))
    >>> list(description_x_acceleration.columns.keys())
    ['counter', 'timestamp', 'x']

    """

    description_class = type(
        "ExtendedAccelerationDescription",
        (AccelerationDescription,),
        attributes,
    )
    return description_class


# -- Classes ------------------------------------------------------------------


class AccelerationDescription(IsDescription):
    """Description of HDF acceleration table"""

    counter = UInt8Col()
    timestamp = UInt64Col()  # Microseconds since measurement start


class StorageException(Exception):
    """Exception for HDF storage errors"""


class Storage:
    """Code to store measurement data in HDF5 format"""

    def __init__(self, filepath: Union[Path, str]) -> None:
        """Initialize the storage object using the given arguments

        Parameters
        ----------

        filepath:
            The filepath of the HDF5 file in which this object should store
            measurement data

        Example
        -------

        >>> filepath = Path("test.hdf5")
        >>> with Storage(filepath) as storage:
        ...     pass
        >>> filepath.unlink()

        """

        self.filepath = Path(filepath).expanduser().resolve()

        self.hdf: Optional[File] = None
        self.acceleration: Optional[Node] = None
        self.start_time: Optional[float] = None

    def __enter__(self) -> Storage:
        """Open the HDF file for writing"""

        self.open()
        return self

    def __exit__(
        self,
        exception_type: Optional[Type[BaseException]],
        exception_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Clean up the resources used by the storage class

        Parameters
        ----------

        exception_type:
            The type of the exception in case of an exception

        exception_value:
            The value of the exception in case of an exception

        traceback:
            The traceback in case of an exception

        """

        self.close()

    def open(self) -> None:
        """Open and initialize the HDF file for writing"""

        try:
            self.hdf = open_file(
                self.filepath,
                mode="w",
                filters=Filters(4, "zlib"),
                title="STH Measurement Data",
            )
        except (HDF5ExtError, OSError) as error:
            raise StorageException(
                f"Unable to open file “{self.filepath}”: {error}"
            )

    def init_acceleration(
        self, axes: Iterable[str], start_time: float
    ) -> None:
        """Initialize the data storage for the collection of acceleration data

        Parameters
        ----------

        axes:
            All acceleration axes for which data should be collected

        start_time:
            The start time of the data acquisition in milliseconds

        Examples
        --------

        >>> filepath = Path("test.hdf5")
        >>> with Storage(filepath) as storage:
        ...      storage.init_acceleration(axes = ['x'], start_time = 1337.42)
        >>> filepath.unlink()

        >>> with Storage(filepath) as storage:
        ...      storage.init_acceleration(axes = ['x', 'z'],
        ...                                start_time = 42.1337)
        >>> filepath.unlink()

        """

        if self.hdf is None:
            self.open()

        assert isinstance(self.hdf, File)

        name = "acceleration"
        try:
            self.acceleration = self.hdf.get_node(f"/{name}")
        except NoSuchNodeError:
            self.acceleration = self.hdf.create_table(
                self.hdf.root,
                name=name,
                description=create_acceleration_description(
                    attributes={axis: Float32Col() for axis in axes}
                ),
                title="STH Acceleration Data",
            )

        self.start_time = start_time
        self.acceleration.attrs["Start_Time"] = datetime.now().isoformat()

    def add_acceleration(
        self, values: Dict[str, float], counter: int, timestamp: float
    ) -> None:
        """Append acceleration data

        Parameters
        ----------

        values:
            The acceleration values that should be added

            - The key specifies the acceleration axis e.g. `x`, `y` or `z`
            - The value specifies the acceleration in the given direction

        counter:
            The message counter sent in the package that contained the
            acceleration value

        timestamp:
            The timestamp of the acceleration message in milliseconds

        Example
        -------

        >>> filepath = Path("test.hdf5")
        >>> with Storage(filepath) as storage:
        ...     storage.add_acceleration(values={'x': 12}, counter=1,
        ...                                    timestamp=4306978.449)
        >>> filepath.unlink()

        """

        if self.acceleration is None:
            self.init_acceleration(values.keys(), timestamp)

        assert isinstance(self.acceleration, Node)
        assert isinstance(self.start_time, (int, float))

        row = self.acceleration.row
        timestamp = (timestamp - self.start_time) * 1000
        row["timestamp"] = timestamp
        row["counter"] = counter
        for accelertation_type, value in values.items():
            row[accelertation_type] = value
        row.append()

        # Flush data to disk every few values to keep memory usage in check
        if self.acceleration.nrows % 1000 == 0:
            self.acceleration.flush()

    def add_streaming_data(self, streaming_data: StreamingData) -> None:
        """Add streaming data to the storage object

        Parameters
        ----------

        streaming_data:
            The streaming data that should be added to the storage

        Examples
        --------

        >>> from mytoolit.measurement import g0
        >>> from mytoolit.can.streaming import TimestampedValue

        >>> t1 = TimestampedValue(timestamp=1, value=g0(1), counter=21)
        >>> t2 = TimestampedValue(timestamp=2, value=g0(2), counter=22)

        >>> t3 = TimestampedValue(timestamp=1, value=g0(3), counter=21)
        >>> t4 = TimestampedValue(timestamp=2, value=g0(4), counter=22)

        >>> data = StreamingData(first=[t1, t2], third=[t3, t4])
        >>> filepath = Path("test.hdf5")
        >>> with Storage(filepath) as storage:
        ...     storage.add_streaming_data(data)
        >>> filepath.unlink()

        """

        # Data for each channel must have same length otherwise (some) data
        # will not be stored.
        assert (
            len(
                {
                    len(channel)
                    for channel in streaming_data
                    if len(channel) != 0
                }
            )
            == 1
        ), "Heterogenous channel lengths"

        axes = []
        data = []
        length_data = 0
        for axis, channel in zip("xyz", streaming_data):
            if len(channel) > 0:
                axes.append(axis)
                data.append(channel)

        if not data:
            return

        if self.acceleration is None:
            self.init_acceleration(axes, data[0][0].timestamp)

        assert isinstance(self.acceleration, Node)
        assert isinstance(self.start_time, (int, float))

        row = self.acceleration.row
        length_data = len(data[0])
        for data_index in range(length_data):
            first = data[0][data_index]

            row["timestamp"] = (first.timestamp - self.start_time) * 1_000_000
            row["counter"] = first.counter
            for index, axis in enumerate(axes):
                value = data[index][data_index].value
                row[axis] = (
                    value.magnitude if isinstance(value, Quantity) else value
                )

            row.append()

        # Flush data to disk every few values to keep memory usage in check
        if self.acceleration.nrows % 1000 == 0:
            self.acceleration.flush()

    def add_acceleration_meta(self, name: str, value: str) -> None:
        """Add acceleration metadata

        Precondition
        ------------

        Either the method

        - init_acceleration or
        - add_acceleration

        have to be called once before you use this method. This is required
        since otherwise the table that stores the acceleration (meta) data
        does not exist.

        Parameters
        ----------

        name:
            The name of the meta attribute

        value:
            The value of the meta attribute

        Example
        -------

        >>> filepath = Path("test.hdf5")
        >>> with Storage(filepath) as storage:
        ...     storage.add_acceleration(values={'x': 12}, counter=1,
        ...                                    timestamp=4306978.449)
        ...     storage.add_acceleration_meta('Sensor_Range', "± 100 g₀")
        >>> filepath.unlink()

        """

        if self.acceleration is None:
            raise UserWarning(
                "Unable to add metadata to non existent "
                "acceleration table.\n"
                "Please call either `init_acceleration` or `add_acceleration` "
                "before you use this function"
            )

        self.acceleration.attrs[name] = value

    def close(self) -> None:
        """Close the HDF file"""

        if isinstance(self.hdf, File) and self.hdf.isopen:
            self.hdf.close()


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



# -- Imports ------------------------------------------------------------------

from enum import Enum
from unittest import TextTestResult

# -- Class --------------------------------------------------------------------


class ExtendedTestResult(TextTestResult):
    """Store data about the result of a test"""

    class TestInformation(object):
        """Store additional data of a test result

        We use this class to store test information in a PDF report.
        """

        class Status(Enum):
            """Store the status of a test"""

            success = 0
            failure = 1
            error = 2

        def __init__(self):
            """Initialize a new test info object"""

            self.status = type(self).Status.success
            self.message = ""

        def set_error(self, message):
            """Set the status of the test to error

            Parameters
            ----------

            message:
                Specifies the error message
            """

            self.status = type(self).Status.error
            self.message = message

        def set_failure(self, message):
            """Set the status of the test to failure

            Parameters
            ----------

            message:
                Specifies the failure message
            """

            self.status = type(self).Status.failure
            self.message = message

        def set_success(self):
            """Set the status of the test to success"""

            self.status = type(self).Status.success
            self.message = ""

        def error(self):
            """Check if there was an error

            Returns
            -------

            True if there was an error, False otherwise
            """

            return self.status == type(self).Status.error

        def failure(self):
            """Check if there test failed

            Returns
            -------

            True if the test failed, False otherwise
            """

            return self.status == type(self).Status.failure

    def __init__(self, *arguments, **keyword_arguments):
        """Initialize the test result"""

        super().__init__(*arguments, **keyword_arguments)

        self.last_test = ExtendedTestResult.TestInformation()

    def addFailure(self, test, error):
        """Add information about the latest failure

        Parameters
        ----------

        test:
            The test case that produced the failure

        error:
            A tuple of the form returned by `sys.exc_info()`:
            (type, value, traceback)
        """

        super().addFailure(test, error)

        # Store message for latest failure
        failure_message = str(error[1])
        # Only store custom message added to assertion, since it should be more
        # readable for a person. If there was no custom message, then the
        # object stores the auto-generated message.
        custom_failure_message = failure_message.split(" : ")[-1]

        self.last_test.set_failure(custom_failure_message)

    def addError(self, test, error):
        """Add information about the latest error

        This should usually not happen unless there are problems with the
        connection or the syntax of the current code base.

        Parameters
        ----------

        test:
            The test case that produced the error

        error:
            A tuple of the form returned by `sys.exc_info()`:
            (type, value, traceback)
        """

        super().addError(test, error)

        self.last_test.set_error(error[1])

    def addSuccess(self, test):
        """Add information about latest successful test

        Parameters
        ----------

        test:
            The successful test
        """

        super().addSuccess(test)

        self.last_test.set_success()



# -- Import -------------------------------------------------------------------

from datetime import date
from functools import partial
from importlib.resources import as_file, files
from numbers import Real
from os import makedirs
from pathlib import Path
from platform import system
from sys import stderr
from typing import List, Optional

from dynaconf import Dynaconf, ValidationError, Validator
from platformdirs import site_config_dir, user_config_dir

from mytoolit.utility.open import open_file, UnableToOpenError

# -- Classes ------------------------------------------------------------------


class ConfigurationUtility:
    """Access configuration data"""

    app_name = "ICOc"
    app_author = "MyTooliT"
    config_filename = "config.yaml"
    site_config_filepath = (
        Path(site_config_dir(app_name, appauthor=app_author)) / config_filename
    )
    user_config_filepath = (
        Path(user_config_dir(app_name, appauthor=app_author)) / config_filename
    )

    @staticmethod
    def open_config_file(filepath: Path):
        """Open configuration file

        Parameters
        ----------

        filepath:
            Path to configuration file

        """

        # Create file, if it does not exist already
        if not filepath.exists():
            filepath.parent.mkdir(
                exist_ok=True,
                parents=True,
            )

            default_user_config = (
                files("mytoolit.config").joinpath("user.yaml").read_text()
            )

            with filepath.open("w", encoding="utf8") as config_file:
                config_file.write(default_user_config)

        open_file(filepath)

    @classmethod
    def open_user_config(cls):
        """Open the current users configuration file"""

        try:
            cls.open_config_file(cls.user_config_filepath)
        except UnableToOpenError as error:
            print(
                (
                    f"Unable to open user configuration: {error}"
                    "\nTo work around this problem please open "
                    f"“{cls.user_config_filepath}” in your favorite text "
                    "editor"
                ),
                file=stderr,
            )


class SettingsIncorrectError(Exception):
    """Raised when the configuration is incorrect"""


class Settings(Dynaconf):
    """Small extension of the settings object for our purposes"""

    def __init__(
        self,
        default_settings_filepath,
        settings_files: Optional[List[str]] = None,
        *arguments,
        **keyword_arguments,
    ) -> None:
        """Initialize the settings using the given arguments

        Parameters
        ----------

        default_settings_filepath:
            Filepath to default settings file

        setting_files:
            A list containing setting files in ascending order according to
            importance (most important last).

        arguments:
            All positional arguments

        keyword_arguments:
            All keyword arguments

        """

        if settings_files is None:
            settings_files = []

        settings_files = [
            default_settings_filepath,
            ConfigurationUtility.site_config_filepath,
            ConfigurationUtility.user_config_filepath,
        ] + settings_files

        super().__init__(
            settings_files=settings_files,
            *arguments,
            **keyword_arguments,
        )
        self.validate_settings()

    def validate_settings(self) -> None:
        """Check settings for errors"""

        def MustExist(must_exist, *arguments, **keyword_arguments):
            """Return Validator which requires setting to exist"""

            return Validator(*arguments, must_exist=True, **keyword_arguments)

        def element_is_type(
            nodes,
            name: str,
            element_type: type,
        ):
            """Check that all elements of a list have a certain type"""
            if nodes is None:
                return True  # Let parent validator handle wrong type

            for node in nodes:
                if not isinstance(node, element_type):
                    raise ValidationError(
                        f"Element “{node}” of {name} has wrong type "
                        f"“{type(node)}” instead of “{element_type}”"
                    )
            return True

        def element_is_string(nodes, name: str):
            """Check that all elements of a list are strings"""
            return element_is_type(nodes, name, element_type=str)

        def element_is_int(nodes, name: str):
            """Check that all elements of a list are ints"""
            return element_is_type(nodes, name, element_type=int)

        def device_validators(name: str):
            """Return shared validator for ICOtronic device (STH, STU, SMH)"""

            return [
                MustExist(
                    f"{name}.batch_number",
                    f"{name}.gtin",
                    f"{name}.programming_board.serial_number",
                    is_type_of=int,
                ),
                MustExist(
                    f"{name}.firmware.location.flash",
                    f"{name}.firmware.release_name",
                    f"{name}.hardware_version",
                    is_type_of=str,
                ),
                MustExist(
                    f"{name}.oem_data",
                    is_type_of=list,
                    condition=partial(element_is_int, name="{name}.oem_data"),
                ),
                MustExist(
                    f"{name}.production_date",
                    is_type_of=date,
                ),
                MustExist(
                    f"{name}.product_name",
                    is_type_of=str,
                    len_max=128,
                    cast=str,
                ),
                MustExist(
                    f"{name}.serial_number",
                    is_type_of=str,
                    len_max=8,
                    cast=str,
                ),
            ]

        def sensor_device_validators(name: str):
            """Return shared validator for STH or SMH"""

            return device_validators(name) + [
                MustExist(
                    f"{name}.name",
                    is_type_of=str,
                ),
            ]

        def sensor_validators(name: str):
            prefix = "sth.acceleration_sensor"
            return [
                MustExist(
                    f"{prefix}.{name}.acceleration.maximum",
                    f"{prefix}.{name}.acceleration.ratio_noise_to_max_value",
                    f"{prefix}.{name}.acceleration.tolerance",
                    f"{prefix}.{name}.reference_voltage",
                    f"{prefix}.{name}.self_test.voltage.difference",
                    f"{prefix}.{name}.self_test.voltage.tolerance",
                    is_type_of=Real,
                ),
                MustExist(
                    f"{prefix}.{name}.self_test.dimension",
                    is_type_of=str,
                    is_in=("x", "y", "z"),
                ),
            ]

        config_system = "mac" if system() == "Darwin" else system().lower()
        can_validators = [
            MustExist(f"can.{config_system}.bitrate", is_type_of=int),
            MustExist(
                f"can.{config_system}.channel",
                f"can.{config_system}.interface",
                is_type_of=str,
            ),
        ]
        commands_validators = [
            MustExist(
                "commands.path.linux",
                is_type_of=list,
                condition=partial(
                    element_is_string, name="commands.path.linux"
                ),
            ),
            MustExist(
                "commands.path.mac",
                is_type_of=list,
                condition=partial(element_is_string, name="commands.path.mac"),
            ),
            MustExist(
                "commands.path.windows",
                is_type_of=list,
                condition=partial(
                    element_is_string, name="commands.path.windows"
                ),
            ),
        ]
        logger_validators = [
            MustExist(
                "logger.can.level",
                is_type_of=str,
                is_in=(
                    "CRITICAL",
                    "ERROR",
                    "WARNING",
                    "INFO",
                    "DEBUG",
                    "NOTSET",
                ),
            )
        ]
        gui_validators = [
            MustExist("gui.host", is_type_of=str),
            MustExist("gui.port", is_type_of=int),
        ]
        measurement_validators = [
            MustExist(
                "measurement.output.directory",
                "measurement.output.filename",
                is_type_of=str,
            ),
        ]
        operator_validators = [MustExist("operator.name", is_type_of=str)]
        sensory_device_validators = [
            MustExist(
                "sensory_device.bluetooth.advertisement_time_1",
                "sensory_device.bluetooth.advertisement_time_2",
                "sensory_device.bluetooth.sleep_time_1",
                "sensory_device.bluetooth.sleep_time_2",
                is_type_of=int,
            )
        ]
        smh_validators = sensor_device_validators("smh") + [
            MustExist(
                "smh.channels",
                is_type_of=int,
            )
        ]
        sth_validators = (
            sensor_device_validators("sth")
            + sensor_validators("ADXL1001")
            + sensor_validators("ADXL1002")
            + sensor_validators("ADXL356")
        ) + [
            MustExist(
                "sth.acceleration_sensor.sensor",
                is_in=(
                    "ADXL1001",
                    "ADXL1002",
                    "ADXL356",
                ),
            ),
            MustExist(
                "sth.battery_voltage.average",
                "sth.battery_voltage.tolerance",
                is_type_of=Real,
            ),
            MustExist(
                "sth.status",
                is_in=(
                    "Bare PCB",
                    "Epoxied",
                ),
            ),
        ]
        stu_validators = device_validators("stu")

        self.validators.register(
            *can_validators,
            *commands_validators,
            *logger_validators,
            *gui_validators,
            *measurement_validators,
            *operator_validators,
            *sensory_device_validators,
            *smh_validators,
            *sth_validators,
            *stu_validators,
        )

        try:
            self.validators.validate()
        except ValidationError as error:
            config_files_text = "\n".join(
                (
                    f"  • {ConfigurationUtility.site_config_filepath}",
                    f"  • {ConfigurationUtility.user_config_filepath}",
                )
            )
            raise SettingsIncorrectError(
                f"Incorrect configuration: {error}\n\n"
                "Please make sure that the configuration files:\n\n"
                f"{config_files_text}\n\n"
                "contain the correct configuration values"
            ) from error

    def acceleration_sensor(self):
        """Get the settings for the current acceleration sensor

        Returns
        -------

        A configuration object for the currently selected accelerometer sensor
        """

        sensor_settings = self.sth.acceleration_sensor
        if sensor_settings.sensor == "ADXL1001":
            return sensor_settings.adxl1001
        if sensor_settings.sensor == "ADXL1002":
            return sensor_settings.adxl1002
        if sensor_settings.sensor == "ADXL356":
            return sensor_settings.adxl356

        # TODO: Use [validation](https://www.dynaconf.com/validation)
        # to handle incorrect config values after
        #
        # - https://github.com/rochacbruno/dynaconf/issues/486
        #
        # is fixed.
        raise ValueError(
            f"Unsupported sensor: “{sensor_settings.sensor}”\n\n"
            "Please use one of the supported sensor configuration values "
            "“ADXL1001” or “ADXL1002” or “ADXL356”"
        )

    def sth_name(self) -> str:
        """Return the current name of the STH as string

        Returns
        -------

        The name of the STH

        """

        return str(self.sth.name)

    def output_directory(self) -> Path:
        """Get the HDF output directory

        Returns
        -------

        The HDF output directory as path object

        """

        directory = Path(settings.measurement.output.directory)
        return directory if directory.is_absolute() else directory.expanduser()

    def check_output_directory(self) -> None:
        """Check the output directory

        If the directory does not already exist, then this function will try to
        create it.

        """

        directory = self.output_directory()

        if directory.exists() and not directory.is_dir():
            raise NotADirectoryError(
                f"The output directory “{directory}” points to an "
                "existing file not a directory"
            )

        if not directory.is_dir():
            try:
                makedirs(str(directory))
            except OSError as error:
                raise OSError(
                    "Unable to create the output directory "
                    f"“{directory}”: {error}"
                )


# -- Attributes ---------------------------------------------------------------


with as_file(
    files("mytoolit.config").joinpath(ConfigurationUtility.config_filename)
) as repo_settings_filepath:
    try:
        settings = Settings(default_settings_filepath=repo_settings_filepath)
    except SettingsIncorrectError as error:
        print(f"{error}", file=stderr)
        exit(1)



# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from typing import List, Optional, Union

from can.interfaces.pcan.basic import PCAN_MESSAGE_EXTENDED, TPCANMsg
from can import Message as CANMessage
from netaddr import EUI

from mytoolit.can.calibration import CalibrationMeasurementFormat
from mytoolit.can.command import Command
from mytoolit.can.identifier import Identifier
from mytoolit.can.node import Node
from mytoolit.can.status import State
from mytoolit.can.streaming import StreamingFormat, StreamingFormatVoltage
from mytoolit.utility import convert_bytes_to_text

# -- Class --------------------------------------------------------------------


class Message:
    """Wrapper class for CAN messages"""

    def __init__(
        self,
        *message: Union[TPCANMsg, CANMessage],
        identifier: Optional[Identifier] = None,
        data: Optional[List[int]] = None,
        **keyword_arguments: Union[Command, Node, None, str, int, bool],
    ) -> None:
        """Create a new message based on the given attributes

        Usually you will either specify the (PCAN or python-can) message
        directly, or provide identifier and data. If, you decide to specify
        both the message and one of the keyword arguments, then the keyword
        arguments will be used to overwrite specific parts of the message
        argument.

        Additional keyword parameters will be used as arguments for the
        identifier. This way you can specify all the arguments for a message
        without creating the identifier class yourself. If you decide to
        specify both an identifier and keyword arguments for the identifier,
        then the keyword arguments will overwrite the specific parts of the
        provided identifier.

        Parameters
        ----------

        message:
            Either
                 - A PCAN message structure as used by the PCAN Basic API
                 - or a python-can message

        identifier:
            A 29 bit CAN identifier

        data:
            An iterable over 8 bit values that stores the payload of the
            message

        Examples
        --------

        Create a message from a given PCAN message

        >>> message = Message(TPCANMsg())

        Create a message from a python-can message

        >>> from can import Message as CANMessage
        >>> message = Message(
        ...     CANMessage(
        ...         is_extended_id=True,
        ...         arbitration_id=0b0_000000_00000001_00_0_01111_0_00001))
        >>> from re import search
        >>> search('# (.*)', repr(message))[1]
        '[SPU 1 → STH 1, Block: System, Command: Reset, Acknowledge]'

        Create a message using identifier and data

        >>> identifier = Identifier(block=0, block_command=1)
        >>> payload = [0xab, 0xcd]
        >>> message = Message(identifier=identifier, data=payload)

        Create a message using keyword arguments handled by the identifier
        class

        >>> message = Message(block='System', block_command='Reset')

        If you provide both identifier and identifier keyword arguments, then
        the specific keyword arguments overwrite the values of the identifier

        >>> identifier = Identifier(block='System')
        >>> message = Message(identifier=identifier, block='Streaming')
        >>> Identifier(message.id()).block_name()
        'Streaming'

        """

        if message:
            can_message = message[0]
            if isinstance(can_message, TPCANMsg):
                self.message = CANMessage(
                    is_extended_id=True,
                    arbitration_id=can_message.ID,
                    data=[
                        can_message.DATA[byte]
                        for byte in range(can_message.LEN)
                    ],
                )
            elif isinstance(can_message, CANMessage):
                self.message = can_message
            else:
                raise ValueError(
                    "Unsupported object type for argument message: "
                    f"“{type(can_message)}”"
                )
        else:
            self.message = CANMessage(is_extended_id=True)

        if identifier:
            self.message.arbitration_id = (
                identifier if isinstance(identifier, int) else identifier.value
            )

        if keyword_arguments:
            # mypy assumes that all keyword arguments have the same type
            self.message.arbitration_id = Identifier(
                self.id(), **keyword_arguments  # type: ignore
            ).value

        if data:
            self.message.data = bytearray(data)
            self.message.dlc = len(data)

    @property
    def data(self) -> bytearray:
        """Retrieve the message data

        Returns
        -------

        A bytearray containing the data of the message

        Examples
        --------

        >>> len(Message().data)
        0
        >>> len(Message(data=[5, 6, 7]).data)
        3
        >>> len(Message(data=[1, 2]).data)
        2

        """

        return self.message.data

    @data.setter
    def data(self, data: Union[List[int], bytearray]) -> None:
        """Set the message data to a new value

        Parameters
        ----------

        data:
            The data that should be stored in the message

        Examples
        --------

        >>> message = Message(data=[1])
        >>> message.data[0] = 42
        >>> message.data[0]
        42

        >>> message = Message(data = bytearray([0] * 8))
        >>> message.data[1] = 1
        >>> message.data[2] = 2
        >>> message.data[7] = 7

        >>> message.data[1]
        1
        >>> message.data[3]
        0
        >>> message.data[7]
        7

        """

        self.message.data = bytearray(data)
        self.message.dlc = len(self.data)

    def _data_explanation_system(self) -> str:
        """Retrieve a textual representation of system messages

        Returns
        -------

        A textual description of the data contained in the message

        """

        def mac_address() -> EUI:
            """Convert the message data into a MAC address"""
            return EUI("-".join((f"{byte:0x}" for byte in self.data[7:1:-1])))

        identifier = self.identifier()
        data_explanation = ""

        if identifier.block_command_name() == "Get/Set State":
            if len(self.data) < 1:
                return ""

            data_explanation = repr(State(self.data[0]))

        if identifier.block_command_name() == "Bluetooth":
            subcommand = self.data[0]
            device_number = self.data[1]
            is_acknowledgment = self.identifier().is_acknowledgment()
            verb = "Return" if is_acknowledgment else "Get"

            if subcommand == 1:
                verb = "Acknowledge" if is_acknowledgment else "Request"
                data_explanation = f"{verb} Bluetooth activation"
            elif subcommand == 2:
                data_explanation = f"{verb} number of available devices"
                if is_acknowledgment:
                    number_devices_text = convert_bytes_to_text(self.data[2:])
                    data_explanation += ": "
                    try:
                        number_devices = int(number_devices_text)
                        data_explanation += f"{number_devices}"
                    except ValueError:
                        data_explanation += (
                            "Unable to convert text "
                            f"“{number_devices_text}” to number"
                        )

            elif subcommand == 5 or subcommand == 6:
                part = "first" if subcommand == 5 else "second"
                data_explanation = (
                    f"{verb} {part} part of name of device "
                    f"with device number “{device_number}”"
                )
                if is_acknowledgment and len(self.data) >= 2:
                    name = convert_bytes_to_text(self.data[2:])
                    data_explanation += f": “{name}”"
            elif subcommand == 7:
                info = (
                    "Acknowledge connection request"
                    if is_acknowledgment
                    else "Request connection"
                )
                data_explanation = (
                    f"{info} to device with device number “{device_number}”"
                )
            elif subcommand == 8:
                data_explanation = f"{verb} Bluetooth connection status"
                if is_acknowledgment and len(self.data) >= 3:
                    connected = bool(self.data[2])
                    data_explanation += ": {}".format(
                        "Connected" if connected else "Not connected"
                    )
            elif subcommand == 9:
                verb = "Acknowledge" if is_acknowledgment else "Request"
                data_explanation = f"{verb} Bluetooth deactivation"
            elif subcommand == 10:
                verb = "Return" if is_acknowledgment else "Get"
                data_explanation = f"{verb} Bluetooth send counter"
                if is_acknowledgment and len(self.data) >= 8:
                    counter = int.from_bytes(self.data[2:], byteorder="big")
                    data_explanation += f": {counter}"
            elif subcommand == 12:
                data_explanation = (
                    f"{verb} RSSI of device "
                    f"with device number “{device_number}”"
                )
                if is_acknowledgment and len(self.data) >= 3:
                    rssi = int.from_bytes(
                        self.data[2:3], byteorder="little", signed=True
                    )
                    data_explanation += f": {rssi}"
            elif subcommand == 14:
                data_explanation = "Write energy mode reduced"
                if len(self.data) >= 8:
                    time_normal_to_reduced_ms = int.from_bytes(
                        self.data[2:6], byteorder="little"
                    )
                    ADVERTISEMENT_TIME_EEPROM_TO_MS = 0.625
                    advertisement_time = (
                        int.from_bytes(self.data[6:], byteorder="little")
                        * ADVERTISEMENT_TIME_EEPROM_TO_MS
                    )
                    data_explanation += ": " + ", ".join(
                        [
                            f"⟳ {time_normal_to_reduced_ms} ms",
                            f"📢 {advertisement_time} ms",
                        ]
                    )
            elif subcommand == 17:
                data_explanation = (
                    f"{verb} MAC address of device "
                    f"with device number “{device_number}”"
                )
                if is_acknowledgment and len(self.data) >= 8:
                    data_explanation += f": {mac_address()}"
            elif subcommand == 18:
                verb = (
                    "Acknowledge connection request"
                    if is_acknowledgment
                    else "Request connection"
                )
                data_explanation = (
                    f"{verb} to device with MAC address “{mac_address()}”"
                )

        return data_explanation

    def _data_explanation_streaming(self) -> str:
        """Retrieve a textual representation of streaming messages

        Returns
        -------

        A textual description of the data contained in the message

        """

        if len(self.data) < 1:
            return ""

        identifier = self.identifier()
        data_explanation = ""

        block_command = identifier.block_command_name()

        StreamingFormatClass = (
            StreamingFormat
            if block_command == "Data"
            else (
                StreamingFormatVoltage
                if block_command == "Voltage"
                else StreamingFormat
            )
        )
        streaming_format: StreamingFormat = StreamingFormatClass(self.data[0])
        data_explanation += repr(streaming_format)

        if identifier.is_acknowledgment() and len(self.data) >= 2:
            sequence_counter = self.data[1]
            explanations = [f"Sequence Counter: {sequence_counter}"]
            size_value = streaming_format.data_bytes()
            values = [
                int.from_bytes(
                    self.data[start : start + size_value], byteorder="little"
                )
                for start in range(2, len(self.data), size_value)
            ]
            explanations.extend(
                [
                    f"{explanation}: {value}"
                    for explanation, value in zip(
                        streaming_format.value_explanations, values
                    )
                ]
            )
            data_explanation += ", " + ", ".join(explanations)

        return data_explanation

    def _data_explanation_configuration(self) -> str:
        identifier = self.identifier()
        data_explanation = ""

        if (
            identifier.block_command_name() == "Calibration Measurement"
            and len(self.data) >= 4
        ):
            return repr(CalibrationMeasurementFormat(self.data))

        return data_explanation

    def _data_explanation_eeprom(self) -> str:
        """Retrieve a textual representation of EEPROM messages

        Returns
        -------

        A textual description of the data contained in the message

        """

        identifier = self.identifier()
        data_explanation = ""

        if identifier.block_command_name() == "Read Write Request Counter":
            counter = int.from_bytes(self.data[4:], "little")
            data_explanation = f"EEPROM Write Requests: {counter}"
        elif (
            identifier.block_command_name() in {"Read", "Write"}
            and len(self.data) >= 8
        ):
            page = self.data[0]
            offset = self.data[1]
            length = self.data[2]
            data = list(self.data[4 : 4 + min(length, 4)])
            is_acknowledgment = self.identifier().is_acknowledgment()

            info = "Acknowledge request" if is_acknowledgment else "Request"
            verb = identifier.block_command_name().lower()
            data_explanation = (
                f"{info} to {verb} {length} bytes at page {page} "
                f"with offset {offset}"
            )
            read_request = verb == "read" and not is_acknowledgment
            if not read_request:
                data_explanation += f": {data}"

        return data_explanation

    def _data_explanation(self) -> str:
        """Retrieve a textual representation of the message data

        Returns
        -------

        A textual description of the data contained in the message

        """

        if len(self.data) <= 0:
            return ""

        identifier = self.identifier()

        if identifier.block_name() == "System":
            return self._data_explanation_system()
        elif identifier.block_name() == "Streaming":
            return self._data_explanation_streaming()
        elif identifier.block_name() == "Configuration":
            return self._data_explanation_configuration()
        elif identifier.block_name() == "EEPROM":
            return self._data_explanation_eeprom()

        return ""

    def __repr__(self) -> str:
        """Get a textual representation of the current message

        Returns
        -------

        A text that shows the various attributes of the current message

        Examples
        --------

        >>> pcan_message = TPCANMsg()
        >>> pcan_message.ID = Identifier(block=0, block_command=1,
        ...                             request=True, error=False,
        ...                             sender=1, receiver=14).value
        >>> pcan_message.DATA[0] = 0xfe
        >>> pcan_message.DATA[1] = 0xfe
        >>> pcan_message.LEN = 2
        >>> Message(pcan_message) # doctest:+NORMALIZE_WHITESPACE
        0b00000000000000110000001001110 2 0xfe 0xfe
        # [STH 1 → STH 14, Block: System, Command: Reset, Request]

        >>> from re import search
        >>> representation = repr(
        ...     Message(block='Streaming',
        ...             block_command='Data',
        ...             sender='SPU 1',
        ...             receiver='STH 1',
        ...             request=True))
        >>> search('# (.*)', representation)[1]
        '[SPU 1 → STH 1, Block: Streaming, Command: Data, Request]'

        >>> representation = repr(
        ...     Message(block='System',
        ...             block_command='Bluetooth',
        ...             sender='SPU 1',
        ...             receiver='STU 1',
        ...             request=True,
        ...             data=[1] + [0]*7))
        >>> search('# (.*)', representation)[1] # doctest:+NORMALIZE_WHITESPACE
        '[SPU 1 → STU 1, Block: System, Command: Bluetooth, Request]
         (Request Bluetooth activation)'

        >>> representation = repr(
        ...     Message(block='System',
        ...             block_command='Get/Set State',
        ...             sender='SPU 2',
        ...             receiver='STU 1',
        ...             request=True,
        ...             data=[0b10_0_101]))
        >>> search('# (.*)', representation)[1] # doctest:+NORMALIZE_WHITESPACE
        '[SPU 2 → STU 1, Block: System, Command: Get/Set State, Request]
         (Get State, Location: Application, State: Operating)'

        """

        identifier = self.identifier()
        data_explanation = self._data_explanation()

        explanation = (
            f"{identifier} ({data_explanation})"
            if data_explanation
            else repr(identifier)
        )

        data_representation = " ".join(
            [hex(self.data[byte]) for byte in range(len(self.data))]
        )
        bit_values = [
            f"0b{identifier.value:029b}",
            str(len(self.data)),
            data_representation,
        ]
        # Filter empty string, since otherwise there might be an additional
        # space at the end of the representation for empty data
        bit_representation = " ".join(filter(None, bit_values))

        return f"{bit_representation} # {explanation}"

    def acknowledge(self, error: bool = False) -> Message:
        """Returns an acknowledgment message object for this message

        In the acknowledgment message receiver and sender will be swapped and
        the request (acknowledge) bit will be set to 0 (acknowledge). The data
        of the acknowledgment message will be the same as in the original
        message.

        Parameters
        ----------

        error:
            Specifies if the acknowledgment message is an error acknowledgment
            or not

        Returns
        -------

        An acknowledgment message for the current message

        Example
        -------

        >>> identifier = Identifier(block=0, block_command=1, sender=5,
        ...                         receiver=10)
        >>> message = Message(identifier=identifier, data=[0xaa])
        >>> message.acknowledge() # doctest:+NORMALIZE_WHITESPACE
        0b00000000000000100001010000101 1 0xaa
        # [STH 10 → STH 5, Block: System, Command: Reset, Acknowledge]

        """

        return Message(
            identifier=self.identifier().acknowledge(),
            error=error,
            data=list(self.data),
        )

    def id(self) -> int:
        """Retrieve the ID of the message

        Returns
        -------

        The 29 bit CAN identifier of the message


        Example
        -------

        >>> Message(block='Configuration').id(
        ...        ) == 0b0_1010_000000000000000000000000
        True

        """

        return self.message.arbitration_id

    def identifier(self) -> Identifier:
        """Retrieve an identifier object for the ID of the message

        Returns
        -------

        An identifier object representing the ID of the message

        Example
        -------

        >>> Message(block='System', block_command='Reset', request=True,
        ...         sender='SPU 1', receiver='STH 2').identifier()
        [SPU 1 → STH 2, Block: System, Command: Reset, Request]

        """

        return Identifier(self.id())

    def to_python_can(self) -> CANMessage:
        """Retrieve a python-can message object for this message

        Returns
        -------

        A message object of the python-can API

        Example
        -------

        >>> message = Message(block='EEPROM', sender='SPU 1', data=[1,2,3])
        >>> can_message = message.to_python_can()
        >>> can_message # doctest:+NORMALIZE_WHITESPACE
        can.Message(timestamp=0.0, arbitration_id=0xf4003c0,
                    is_extended_id=True, dlc=3, data=[0x1, 0x2, 0x3])

        >>> message.id() == can_message.arbitration_id
        True

        """

        return self.message

    def to_pcan(self) -> TPCANMsg:
        """Retrieve a PCAN message object for this message

        Returns
        -------

        A message object of the PCAN Basic API

        Example
        -------

        >>> message = Message(block='System', block_command='Bluetooth')
        >>> pcan_message = message.to_pcan()
        >>> message.id() == pcan_message.ID
        True

        """

        message = TPCANMsg()
        message.ID = self.message.arbitration_id
        for byte, value in enumerate(self.message.data):
            message.DATA[byte] = value
        message.LEN = len(self.data)
        message.MSGTYPE = PCAN_MESSAGE_EXTENDED

        return message


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



# -- Imports ------------------------------------------------------------------

from os import environ, pathsep
from platform import system

from mytoolit.config import settings

# -- Functions ----------------------------------------------------------------


def add_commander_path_to_environment() -> None:
    """Add path to Simplicity Commander (`commander`) to `PATH`

    Example
    -------

    >>> from os import environ
    >>> length_path_before = len(environ['PATH'])
    >>> add_commander_path_to_environment()
    >>> length_path_after = len(environ['PATH'])
    >>> length_path_after > length_path_before
    True

    """

    path = settings.commands.path
    operating_system = system()
    paths = (
        path.linux
        if operating_system == "Linux"
        else path.mac if operating_system == "Darwin" else path.windows
    )
    environ["PATH"] += pathsep + pathsep.join(paths)


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



# -- Imports ------------------------------------------------------------------

from mytoolit.can import Node

from mytoolit.old.network import Network as OldNetwork
from mytoolit.old.MyToolItNetworkNumbers import MyToolItNetworkNr

# -- Class --------------------------------------------------------------------


class Network:
    def __init__(self, sth_name):
        self.sth_name = sth_name

        self.network = OldNetwork(
            sender=MyToolItNetworkNr["SPU1"], receiver=Node("STU 1").value
        )

        self.network.bConnected = False
        self.network.reset_node("STU 1")

    def __enter__(self):
        self.network.bBlueToothConnectPollingName(
            MyToolItNetworkNr["STU1"], self.sth_name, log=False
        )

        return self.network

    def __exit__(self, exception_type, exception_value, traceback):
        self.network.__exit__()


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    sth_name = "Test-SMH"
    with Network(sth_name) as network:
        print(f"Connected to “{sth_name}”\n")

        sensor_configuration = {"x": 9, "y": 10, "z": 11}
        print(
            "Change sensor configuration to: "
            + ", ".join(
                f"{axis}: {sensor}"
                for axis, sensor in sensor_configuration.items()
            )
        )
        network.write_sensor_config(**sensor_configuration)

        sensor_configuration = network.read_sensor_config()
        print(f"Updated sensor configuration: {sensor_configuration}")



import matplotlib.pyplot as plt
import numpy as np
import socket
import json

from logging import FileHandler, Formatter, getLogger
from pathlib import Path
from time import time

cDict = {
    "Run": True,
    "Plot": False,
    "lineNameX": "",
    "lineNameY": "",
    "lineNameZ": "",
    "diagramName": "Signal Quality",
    "sampleInterval": 0,
    "figSizeX": 13,
    "figSizeY": 6,
    "X-Label": "time [s]",
    "Y-Label": "g",
    "timePoints": None,
    "dataBlockSize": 0,
    "xAccPoints": None,
    "yAccPoints": None,
    "zAccPoints": None,
    "Socket": None,
    "Connection": None,
    "TimeOutMs": 1500,
}


def tArray2Binary(array):
    strDict = json.dumps(array)
    strDict.encode("utf-8")
    binary = " ".join(format(ord(letter), "b") for letter in strDict)
    return binary.encode()


def tBinary2Array(tBinary):
    jsn = "".join(chr(int(x, 2)) for x in tBinary.split())
    msg = None
    try:
        msg = json.loads(jsn)
    except:
        pass
    return msg


def vHandeClose(evt):
    global cDict
    cDict["Run"] = False
    cDict["Plot"] = False


def tPlotterInit():
    global cDict

    # this is the call to matplotlib that allows dynamic plotting
    plt.ion()
    fig = plt.figure(figsize=(cDict["figSizeX"], cDict["figSizeY"]))
    ax = fig.add_subplot(111)
    fig.canvas.mpl_connect("close_event", vHandeClose)
    legendHandles = []
    legendName = []
    line1 = None
    line2 = None
    line3 = None
    # create a variable(s) for the line(s) so we can later update it
    if "" != cDict["lineNameX"]:
        (line1,) = ax.plot(
            cDict["timePoints"], cDict["xAccPoints"], alpha=0.8, label="x"
        )
        legendHandles.append(line1)
        legendName.append(cDict["lineNameX"])
    if "" != cDict["lineNameY"]:
        (line2,) = ax.plot(
            cDict["timePoints"], cDict["yAccPoints"], alpha=0.8, label="Y"
        )
        legendHandles.append(line2)
        legendName.append(cDict["lineNameY"])
    if "" != cDict["lineNameZ"]:
        (line3,) = ax.plot(
            cDict["timePoints"], cDict["zAccPoints"], alpha=0.8, label="Z"
        )
        legendHandles.append(line3)
        legendName.append(cDict["lineNameZ"])
    # update plot label/title
    plt.xlabel(cDict["X-Label"])
    plt.ylabel(cDict["Y-Label"])
    plt.legend(legendHandles, legendName)
    plt.title("{}".format(cDict["diagramName"]))
    plt.show()
    return [line1, line2, line3]


def sPloterSocketInit(iSocketPort):
    """
    Init Server Socket such that Clients may connect

    @param iSocketPort Socket Port to be opened

    @return
    """

    global cDict
    cDict["Socket"] = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    logger = getLogger()
    logger.debug("Created socket")
    cDict["Socket"].bind(
        ("", iSocketPort)  # Symbolic name meaning all available interfaces
    )
    logger.debug(f"Bound socket to port {iSocketPort}")
    cDict["Socket"].listen(1)
    logger.debug("Listening for connections")
    cDict["Connection"], _ = cDict["Socket"].accept()
    logger.debug("Accepted connection")


def vPlotterCommand(command, value):
    global cDict
    if command in {
        "dataBlockSize",
        "diagramName",
        "figSizeX",
        "figSizeY",
        "lineNameX",
        "lineNameY",
        "lineNameZ",
        "Plot",
        "Run",
        "sampleInterval",
        "xDim",
    }:
        cDict[command] = value
    if command == "diagramName" and cDict["Plot"]:
        plt.title("{}".format(cDict["diagramName"]))
    if command == "xDim":
        dataPoints = (
            1000 * cDict["dataBlockSize"] * value / cDict["sampleInterval"]
        )
        cDict["timePoints"] = np.linspace(0, value, int(dataPoints))
        cDict["xAccPoints"] = np.linspace(0, 0, int(dataPoints))
        cDict["yAccPoints"] = np.linspace(0, 0, int(dataPoints))
        cDict["zAccPoints"] = np.linspace(0, 0, int(dataPoints))


def vPlotter(iSocketPort, log_level):
    global cDict

    logger = getLogger(__name__)
    logger.setLevel(log_level)
    handler = FileHandler("plotter.log", delay=True)
    handler.setFormatter(
        Formatter("{asctime} {levelname} {name} {message}", style="{")
    )
    logger.addHandler(handler)

    logger.info("Application started")
    sPloterSocketInit(iSocketPort)
    logger.info("Socket Initialized")
    while cDict["Run"] and not cDict["Plot"]:
        cmd = cDict["Connection"].recv(2**10)
        if cmd is None:
            continue
        cmd = tBinary2Array(cmd)
        if cmd is None:
            continue

        sCommand, tValue = cmd
        message = f"Initialization command: {sCommand}; value: {tValue}"
        logger.debug(message)
        logger.info(message)
        vPlotterCommand(sCommand, tValue)
        cDict["Connection"].sendall(tArray2Binary([sCommand, tValue]))
    logger.info("Configuration set")
    [line1, line2, line3] = tPlotterInit()
    logger.debug("Initialization done")
    logger.info("Configured")
    pauseTime = (1 / cDict["sampleInterval"]) / 4
    tLastTick = int(round(time()))
    logger.info("Drawing started")
    logger.debug("Waiting for sensor data")
    while cDict["Run"]:
        cmd = cDict["Connection"].recv(2**16)
        logger.debug("Received data")

        if cmd is None:
            timeout = cDict["TimeOutMs"] < int(round(time())) - tLastTick()
            if timeout:
                logger.error("Client time out")
                cDict["Run"] = False
                break

        tLastTick = int(round(time() * 1000))
        cmd = tBinary2Array(cmd)
        if cmd is None:
            continue

        sCommand, tValue = cmd
        if sCommand == "data":
            block_size = cDict["dataBlockSize"]
            cDict["xAccPoints"] = cDict["xAccPoints"][block_size:]
            cDict["yAccPoints"] = cDict["yAccPoints"][block_size:]
            cDict["zAccPoints"] = cDict["zAccPoints"][block_size:]
            cDict["xAccPoints"] = np.hstack([cDict["xAccPoints"], tValue["X"]])
            cDict["yAccPoints"] = np.hstack([cDict["yAccPoints"], tValue["Y"]])
            cDict["zAccPoints"] = np.hstack([cDict["zAccPoints"], tValue["Z"]])
            [line1, line2, line3] = vlivePlot(
                cDict["xAccPoints"],
                cDict["yAccPoints"],
                cDict["zAccPoints"],
                line1,
                line2,
                line3,
                pauseTime,
            )
        else:
            logger.info(
                f"Execute non-data command: {sCommand}; value: {tValue}"
            )
            vPlotterCommand(sCommand, tValue)

    logger.info("Closing connection ...")
    cDict["Connection"].close()
    logger.info("Connection closed")


def vlivePlot(yX_data, yY_data, yZ_data, line1, line2, line3, pause_time):
    # After the figure, axis, and line are created, we only need to update the
    # y-data
    update_bounds = False

    if None != line1:
        line1.set_ydata(yX_data)
        if (
            np.min(yX_data) <= line1.axes.get_ylim()[0]
            or np.max(yX_data) >= line1.axes.get_ylim()[1]
        ):
            update_bounds = True
    if None != line2:
        line2.set_ydata(yY_data)
        if (
            np.min(yY_data) <= line2.axes.get_ylim()[0]
            or np.max(yY_data) >= line2.axes.get_ylim()[1]
        ):
            update_bounds = True
    if None != line3:
        line3.set_ydata(yZ_data)
        if (
            np.min(yZ_data) <= line3.axes.get_ylim()[0]
            or np.max(yZ_data) >= line3.axes.get_ylim()[1]
        ):
            update_bounds = True

    # Adjust limits if new data goes beyond bounds
    if update_bounds:
        min_bound = min(
            [
                value
                for value in (
                    (
                        np.min(yX_data) - np.std(yX_data)
                        if yX_data is not None
                        else None
                    ),
                    (
                        np.min(yY_data) - np.std(yY_data)
                        if yY_data is not None
                        else None
                    ),
                    (
                        np.min(yZ_data) - np.std(yZ_data)
                        if yZ_data is not None
                        else None
                    ),
                )
                if value is not None
            ]
        )
        max_bound = max(
            [
                value
                for value in (
                    (
                        np.max(yX_data) + np.std(yX_data)
                        if yX_data is not None
                        else None
                    ),
                    (
                        np.max(yY_data) + np.std(yY_data)
                        if yY_data is not None
                        else None
                    ),
                    (
                        np.max(yZ_data) + np.std(yZ_data)
                        if yZ_data is not None
                        else None
                    ),
                )
                if value is not None
            ]
        )
        plt.ylim(min_bound, max_bound)

    # this pauses the data so the figure/axis can catch up - the amount of
    # pause can be altered above
    plt.pause(pause_time)

    # return line so we can update it again in the next iteration
    return [line1, line2, line3]



# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from re import fullmatch
from typing import Union

# -- Class --------------------------------------------------------------------


class Node:
    """This class represents a CAN node of ICOtronic system"""

    def __init__(self, node: Union[int, str, Node] = 0) -> None:
        """Create a new node from the given argument

        A node represents a communication participant, such as a specific STH
        or a specific STU in the ICOtronic system.

        Parameters
        ----------

        node:
            The number that identifies this node

        Examples
        --------

        Create node with default value

        >>> Node().value
        0

        Create nodes with string values

        >>> Node('STH 1')
        STH 1

        >>> Node('STU1')
        STU 1

        >>> Node('SPU 1').value
        15

        Check that we can not use incorrect numbers to initialize a node

        >>> Node('STU 15')
        Traceback (most recent call last):
           ...
        ValueError: Unknown node identifier “STU 15”

        You can create a copy of an already existing node

        >>> Node(Node('STH 2'))
        STH 2

        """

        if isinstance(node, str):
            # Check for broadcast pseudo nodes
            broadcast_match = fullmatch(
                "Broadcast With(?P<no_acknowledgment>out)? Acknowledgment",
                node,
            )
            if broadcast_match:
                self.value = 0 if broadcast_match["no_acknowledgment"] else 31
                return

            # Check normal nodes
            node_match = fullmatch(
                r"(?P<name>S(?:PU|TH|TU)) ?(?P<number>\d{1,2})", node
            )

            if node_match is not None:
                node_name = node_match["name"]
                node_number = int(node_match["number"])

                if node_name == "STH" and 1 <= node_number <= 14:
                    self.value = node_number
                    return

                if node_name == "SPU" and 1 <= node_number <= 2:
                    self.value = node_number + 14
                    return

                if node_name == "STU" and 1 <= node_number <= 14:
                    self.value = node_number + 16
                    return

            raise ValueError(f"Unknown node identifier “{node}”")

        if isinstance(node, Node):
            self.value = node.value
            return

        self.value = node

    def __repr__(self) -> str:
        """Return the string representation of the current node

        Returns
        -------

        A string that describes the node

        Examples
        --------

        >>> Node(0)
        Broadcast With Acknowledgment

        >>> Node(31)
        Broadcast Without Acknowledgment

        >>> Node(10)
        STH 10

        >>> Node(15)
        SPU 1

        >>> Node(18)
        STU 2

        """

        if self.value == 0 or self.value == 31:
            return "Broadcast With{} Acknowledgment".format(
                "" if self.value == 0 else "out"
            )
        if self.is_sth():
            return f"STH {self.value}"

        if 15 <= self.value <= 16:
            return f"SPU {self.value-14}"

        return f"STU {self.value-16}"

    def is_sth(self) -> bool:
        """Check if this node is an STH or not

        Returns
        -------

        `True` if this node is an STH or `False` otherwise

        Examples
        --------

        >>> Node('STH 1').is_sth()
        True

        >>> Node('STU 12').is_sth()
        False

        >>> Node('STH 12').is_sth()
        True

        """

        return 1 <= self.value <= 14

    def is_stu(self) -> bool:
        """Check if this node is an STU or not

        Returns
        -------

        `True` if this node is an STU or `False` otherwise

        Examples
        --------

        >>> Node('STH 7').is_stu()
        False

        >>> Node('STU 14').is_stu()
        True

        >>> Node('SPU 1').is_stu()
        False

        """

        return 17 <= self.value <= 30


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



# -- Imports ------------------------------------------------------------------

from base64 import b64encode, b64decode
from typing import Union

from netaddr import EUI

# -- Functions ----------------------------------------------------------------


def convert_mac_base64(mac: Union[str, EUI]) -> str:
    """Convert a Bluetooth MAC address to a Base64 encoded text

    Parameters
    ----------

    mac:
        The MAC address

    Returns
    -------

    The MAC address as Base64 encoded string

    Example
    -------

    >>> convert_mac_base64("08:6b:d7:01:de:81")
    'CGvXAd6B'

    """

    return b64encode(EUI(mac).packed).decode()


def convert_base64_mac(name: str) -> str:
    """Convert a Base64 encoded MAC address into a readable MAC address

    Parameters
    ----------

    mac:
        A Base64 encoded text that stores a (Bluetooth) MAC address

    Returns
    -------

    A MAC address as Base64 encoded string

    Example
    -------

    >>> convert_base64_mac('CGvXAd6B')
    '08:6b:d7:01:de:81'

    """

    return ":".join([f"{byte:02x}" for byte in b64decode(name)])


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



from ctypes import BigEndianStructure, c_uint8, Union

from bidict import bidict

AdcMax = 0xFFFF


class ActiveStateFlags(BigEndianStructure):
    """Byte format for get/set state data

    bSetState: Set or get state data
        - 0: Get
        - 1: Set

    bReserved: Reserved bit

    u2NodeState: Current state of the node
        - 0: No Change/Init
        - 1: BootLoader
        - 2: Application
        - 3: Reserved

    bReserved2: Reserved bit

    u3NetworkState: Current state of network
        - 0: Failure (No acknowledgement will be sent;
                      Only power on resets this state)
        - 1: Error (No active communication)
        - 2: Turn Off/Standby
        - 3: Graceful degradation level 2
        - 4: Graceful degradation level 1
        - 5: Operating
        - 6: Startup
        - 7: No change

    See also: https://mytoolit.github.io/Documentation/#command:get-set-state
    """

    _fields_ = [
        ("bSetState", c_uint8, 1),
        ("bReserved", c_uint8, 1),
        ("u2NodeState", c_uint8, 2),
        ("bReserved1", c_uint8, 1),
        ("u3NetworkState", c_uint8, 3),
    ]


class ActiveState(Union):
    _fields_ = [("b", ActiveStateFlags), ("asbyte", c_uint8)]


class AtvcFormatFlags(BigEndianStructure):
    """Byte format for streaming data

    bStreaming: Streaming or single request (latest data)
        - 0: Single request
        - 1: Streaming

    bDataSetBytes: Number of bytes per data set
        - 0: 2 Bytes
        - 1: 3 Bytes

    bNumber1: Collect first part of streaming data (e.g. x-axis, temperature 1)
        - 0: Disable data collection
        - 1: Activate data collection

    bNumber2: Collect second part of streaming data (e.g. y-axis,
              temperature 2)
        - 0: Disable data collection
        - 1: Activate data collection

    bNumber3: Collect third part of streaming data (e.g. z-axis,
              temperature 3)
        - 0: Disable data collection
        - 1: Activate data collection

    u3DataSets: Number of collected data sets
        - 0: Stop (stream)
        - 1: 1 Data sets (x, y, z, x-y-z, x-y, x-z or y-z)
        - 2: 3 Data sets
        - 3: 6 Data sets
        - 4: 10 Data sets
        - 5: 15 Data sets
        - 6: 20 Data sets
        - 7: 30 Data sets (x, y, or z)
    """

    _fields_ = [
        ("bStreaming", c_uint8, 1),
        ("bDataSetBytes", c_uint8, 1),
        ("bNumber1", c_uint8, 1),
        ("bNumber2", c_uint8, 1),
        ("bNumber3", c_uint8, 1),
        ("u3DataSets", c_uint8, 3),
    ]


class AtvcFormat(Union):
    _fields_ = [("b", AtvcFormatFlags), ("asbyte", c_uint8)]


class CalibrationMeassurementFormatFlags(BigEndianStructure):
    _fields_ = [
        ("bSet", c_uint8, 1),
        ("u2Action", c_uint8, 2),
        ("bReset", c_uint8, 1),
        ("Reserved", c_uint8, 4),
    ]


class CalibrationMeassurement(Union):
    _fields_ = [("b", CalibrationMeassurementFormatFlags), ("asbyte", c_uint8)]


MyToolItBlock = bidict(
    {
        "System": 0x00,
        "Streaming": 0x04,
        "StatisticalData": 0x08,
        "Configuration": 0x28,
        "EEPROM": 0x3D,
        "Product Data": 0x3E,
        "Test": 0x3F,
    }
)

MyToolItSystem = bidict(
    {
        "Verboten": 0x00,
        "Reset": 0x01,
        "Get/Set State": 0x02,
        "Mode": 0x03,
        "Alarm": 0x04,
        "Node Status": 0x05,
        "Error Status": 0x06,
        "StatusWord2": 0x07,
        "StatusWord3": 0x08,
        "Test": 0x09,
        "Log": 0x0A,
        "Bluetooth": 0x0B,
        "Routing": 0x0C,
    }
)

SystemCommandBlueTooth = {
    "Reserved": 0,
    "Connect": 1,
    "GetNumberAvailableDevices": 2,
    "SetName1": 3,
    "SetName2": 4,
    "GetName1": 5,
    "GetName2": 6,
    "DeviceConnect": 7,
    "DeviceCheckConnected": 8,
    "Disconnect": 9,
    "SendCounter": 10,
    "ReceiveCounter": 11,
    "Rssi": 12,
    "EnergyModeReducedRead": 13,
    "EnergyModeReducedWrite": 14,
    "EnergyModeLowestRead": 15,
    "EnergyModeLowestWrite": 16,
    "MacAddress": 17,
    "DeviceConnectMacAddr": 18,
    "SelfAddressing": 255,
}

BluetoothTime = {
    "Connect": 20,
    "Out": 4 * 5,
    "Disconnect": 2,
    "GetDeviceNumber": 2,
    "DeviceConnect": 6,
    "TestConnect": 4,
}

SystemCommandRouting = {
    "Reserved": 0,
    "SendCounter": 1,
    "SendFailCounter": 2,
    "SendLowLevelByteCounter": 3,
    "ReceiveCounter": 4,
    "ReceiveFailCounter": 5,
    "ReceiveLowLevelByteCounter": 6,
}

MyToolItStreaming = bidict(
    {
        "Data": 0x00,
        "Temperature": 0x01,
        "Voltage": 0x20,
        "Current": 0x40,
    }
)

MyToolItStatData = bidict(
    {
        "PocPof": 0x00,
        "OperatingTime": 0x01,
        "Uvc": 0x02,
        "Wdog": 0x03,
        "ProductionDate": 0x04,
        "MeasurementInterval": 0x40,
        "QuantityInterval": 0x41,
        "Energy": 0x80,
    }
)

MyToolItConfiguration = bidict(
    {
        "Get/Set ADC Configuration": 0x00,
        "Channel Configuration": 0x01,
        "Voltage": 0x20,
        "Current": 0x40,
        "CalibrationFactorK": 0x60,
        "CalibrationFactorD": 0x61,
        "Calibration Measurement": 0x62,
        "Alarm": 0x80,
        "Hmi": 0xC0,
    }
)

MyToolItEeprom = bidict(
    {"Read": 0x00, "Write": 0x01, "Read Write Request Counter": 0x20}
)

MyToolItProductData = bidict(
    {
        "GTIN": 0x00,
        "Hardware Version": 0x01,
        "Firmware Version": 0x02,
        "Release Name": 0x03,
        "Serial Number 1": 0x04,
        "Serial Number 2": 0x05,
        "Serial Number 3": 0x06,
        "Serial Number 4": 0x07,
        "Product Name 1": 0x08,
        "Product Name 2": 0x09,
        "Product Name 3": 0x0A,
        "Product Name 4": 0x0B,
        "Product Name 5": 0x0C,
        "Product Name 6": 0x0D,
        "Product Name 7": 0x0E,
        "Product Name 8": 0x0F,
        "Product Name 9": 0x10,
        "Product Name 10": 0x11,
        "Product Name 11": 0x12,
        "Product Name 12": 0x13,
        "Product Name 13": 0x14,
        "Product Name 14": 0x15,
        "Product Name 15": 0x16,
        "Product Name 16": 0x17,
        "OEM Free Use 1": 0x18,
        "OEM Free Use 2": 0x19,
        "OEM Free Use 3": 0x1A,
        "OEM Free Use 4": 0x1B,
        "OEM Free Use 5": 0x1C,
        "OEM Free Use 6": 0x1D,
        "OEM Free Use 7": 0x1E,
        "OEM Free Use 8": 0x1F,
    }
)

MyToolItTest = bidict(
    {
        "Signal": 0x01,
        "Pfeifferl": 0x69,
    }
)

blocknumber_to_commands = {
    MyToolItBlock["System"]: MyToolItSystem,
    MyToolItBlock["Streaming"]: MyToolItStreaming,
    MyToolItBlock["StatisticalData"]: MyToolItStatData,
    MyToolItBlock["Configuration"]: MyToolItConfiguration,
    MyToolItBlock["EEPROM"]: MyToolItEeprom,
    MyToolItBlock["Product Data"]: MyToolItProductData,
    MyToolItBlock["Test"]: MyToolItTest,
}

CommandBlock = {
    MyToolItBlock["System"]: "Command Block System",
    MyToolItBlock["Streaming"]: "Command Block Streaming",
    MyToolItBlock["StatisticalData"]: "Command Block Statistical Data",
    MyToolItBlock["Configuration"]: "Command Block Configuration",
    MyToolItBlock["EEPROM"]: "Command Block EEPROM",
    MyToolItBlock["Product Data"]: "Command Block Product Data",
    MyToolItBlock["Test"]: "Command Block Test",
}

CommandBlockSystem = {
    MyToolItSystem["Verboten"]: "System Command Verboten",
    MyToolItSystem["Reset"]: "System Command Reset",
    MyToolItSystem["Get/Set State"]: "System Command Get/Set State",
    MyToolItSystem["Mode"]: "System Command Mode",
    MyToolItSystem["Alarm"]: "System Command Alarm",
    MyToolItSystem["Node Status"]: "System Command Status Word0",
    MyToolItSystem["Error Status"]: "System Command Status Word1",
    MyToolItSystem["StatusWord2"]: "System Command Status Word2",
    MyToolItSystem["StatusWord3"]: "System Command Status Word3",
    MyToolItSystem["Test"]: "System Command Test",
    MyToolItSystem["Log"]: "System Command Log",
    MyToolItSystem["Bluetooth"]: "System Command BlueTooth",
    MyToolItSystem["Routing"]: "System Command Routing",
}

CommandBlockStreaming = {
    MyToolItStreaming["Data"]: "Streaming Command Data",
    MyToolItStreaming["Temperature"]: "Streaming Command Temperature",
    MyToolItStreaming["Voltage"]: "Streaming Command Voltage",
    MyToolItStreaming["Current"]: "Streaming Command Current",
}

CommandBlockStatisticalData = {
    MyToolItStatData["PocPof"]: "Statistical Data Command PowerOn/Off Counter",
    MyToolItStatData[
        "OperatingTime"
    ]: "Statistical Data Command Operating Time",
    MyToolItStatData["Uvc"]: "Statistical Data Command Undervoltage Counter",
    MyToolItStatData["Wdog"]: "Statistical Data Command Watchdog Counter",
    MyToolItStatData[
        "ProductionDate"
    ]: "Statistical Data Command Production Date",
    MyToolItStatData[
        "MeasurementInterval"
    ]: "Statistical Data Command Measurement Interval",
    MyToolItStatData[
        "QuantityInterval"
    ]: "Statistical Data Command Quantity Interval",
    MyToolItStatData["Energy"]: "Statistical Data Command Energy",
}

CommandBlockConfiguration = {
    MyToolItConfiguration[
        "Get/Set ADC Configuration"
    ]: "Configuration Command Acceleration Configuration",
    MyToolItConfiguration[
        "Channel Configuration"
    ]: "Configuration Command Channel Configuration",
    MyToolItConfiguration[
        "Voltage"
    ]: "Configuration Command Voltage Configuration",
    MyToolItConfiguration[
        "Current"
    ]: "Configuration Command Current Configuration",
    MyToolItConfiguration[
        "CalibrationFactorK"
    ]: "Configuration Command Calibration Factor K",
    MyToolItConfiguration[
        "CalibrationFactorD"
    ]: "Configuration Command Calibration Factor D",
    MyToolItConfiguration[
        "Calibration Measurement"
    ]: "Configuration Command Calibration Measurement",
    MyToolItConfiguration["Alarm"]: "Configuration Command Alarm",
    MyToolItConfiguration["Hmi"]: "Configuration Command HMI",
}

CommandBlockEeprom = {
    MyToolItEeprom["Read"]: "EEPROM Command Read",
    MyToolItEeprom["Write"]: "EEPROM Command Write",
    MyToolItEeprom[
        "Read Write Request Counter"
    ]: "EEPROM Command Write Request Counter",
}

CommandBlockProductData = {
    MyToolItProductData["GTIN"]: "Product Data Command GTIN",
    MyToolItProductData[
        "Hardware Version"
    ]: "Product Data Command Hardware Version",
    MyToolItProductData[
        "Firmware Version"
    ]: "Product Data Command Firmware Version",
    MyToolItProductData["Release Name"]: "Product Data Command Release Name",
    MyToolItProductData[
        "Serial Number 1"
    ]: "Product Data Command Serial Number 1",
    MyToolItProductData[
        "Serial Number 2"
    ]: "Product Data Command Serial Number 2",
    MyToolItProductData[
        "Serial Number 3"
    ]: "Product Data Command Serial Number 3",
    MyToolItProductData[
        "Serial Number 4"
    ]: "Product Data Command Serial Number 4",
    MyToolItProductData["Product Name 1"]: "Product Data Command Name1",
    MyToolItProductData["Product Name 2"]: "Product Data Command Name2",
    MyToolItProductData["Product Name 3"]: "Product Data Command Name3",
    MyToolItProductData["Product Name 4"]: "Product Data Command Name4",
    MyToolItProductData["Product Name 5"]: "Product Data Command Name5",
    MyToolItProductData["Product Name 6"]: "Product Data Command Name6",
    MyToolItProductData["Product Name 7"]: "Product Data Command Name7",
    MyToolItProductData["Product Name 8"]: "Product Data Command Name8",
    MyToolItProductData["Product Name 9"]: "Product Data Command Name9",
    MyToolItProductData["Product Name 10"]: "Product Data Command Name10",
    MyToolItProductData["Product Name 11"]: "Product Data Command Name11",
    MyToolItProductData["Product Name 12"]: "Product Data Command Name12",
    MyToolItProductData["Product Name 13"]: "Product Data Command Name13",
    MyToolItProductData["Product Name 14"]: "Product Data Command Name14",
    MyToolItProductData["Product Name 15"]: "Product Data Command Name15",
    MyToolItProductData["Product Name 16"]: "Product Data Command Name16",
    MyToolItProductData["OEM Free Use 1"]: "Product Data Command Free Use 1",
    MyToolItProductData["OEM Free Use 2"]: "Product Data Command Free Use 2",
    MyToolItProductData["OEM Free Use 3"]: "Product Data Command Free Use 3",
    MyToolItProductData["OEM Free Use 4"]: "Product Data Command Free Use 4",
    MyToolItProductData["OEM Free Use 5"]: "Product Data Command Free Use 5",
    MyToolItProductData["OEM Free Use 6"]: "Product Data Command Free Use 6",
    MyToolItProductData["OEM Free Use 7"]: "Product Data Command Free Use 7",
    MyToolItProductData["OEM Free Use 8"]: "Product Data Command Free Use 8",
}

CommandBlockTest = {
    MyToolItTest["Signal"]: "Test Command Signal",
    MyToolItTest["Pfeifferl"]: "Test Command Pfeifferl",
}

EepromPage = {
    "SystemConfiguration0": 0,
    "Product Data": 4,
    "Statistics": 5,
    "Calibration0": 8,
}

CalibMeassurementActionNr = {
    "None": 0,
    "Activate": 1,
    "Deactivate": 2,
    "Measure": 3,
}

CalibMeassurementActionName = {
    0: "Calibration Measurement Action - None/Reset",
    1: "Calibration Measurement Action - Activate",
    2: "Calibration Measurement Action - Deactivate",
    3: "Calibration Measurement Action - Measure",
}

CalibMeassurementTypeNr = {
    "Acc": 0,
    "Temp": 1,
    "Voltage": 32,
    "Vss": 96,
    "Avdd": 97,
    "OpvOutput": 99,
}

CalibMeassurementTypeName = {
    CalibMeassurementTypeNr[
        "Acc"
    ]: "Calibration Measurement Type - Acceleration",
    CalibMeassurementTypeNr[
        "Temp"
    ]: "Calibration Measurement Type - Temperature",
    CalibMeassurementTypeNr[
        "Voltage"
    ]: "Calibration Measurement Type - Voltage",
    CalibMeassurementTypeNr[
        "Vss"
    ]: "Calibration Measurement Type - VSS(Ground)",
    CalibMeassurementTypeNr[
        "Avdd"
    ]: "Calibration Measurement Type - AVDD(Analog Supply)",
    CalibMeassurementTypeNr[
        "OpvOutput"
    ]: "Calibration Measurement Type - OPV Output",
}

AdcAcquisitionTime = bidict(
    {
        1: 0,
        2: 1,
        3: 2,
        4: 3,
        8: 4,
        16: 5,
        32: 6,
        64: 7,
        128: 8,
        256: 9,
    }
)

AdcAcquisitionTimeName = {
    AdcAcquisitionTime[1]: "ADC Acquisition Time - 1 Cycle",
    AdcAcquisitionTime[2]: "ADC Acquisition Time - 2 Cycles",
    AdcAcquisitionTime[3]: "ADC Acquisition Time - 3 Cycles",
    AdcAcquisitionTime[4]: "ADC Acquisition Time - 4 Cycles",
    AdcAcquisitionTime[8]: "ADC Acquisition Time - 8 Cycles",
    AdcAcquisitionTime[16]: "ADC Acquisition Time - 16 Cycles",
    AdcAcquisitionTime[32]: "ADC Acquisition Time - 32 Cycles",
    AdcAcquisitionTime[64]: "ADC Acquisition Time - 64 Cycles",
    AdcAcquisitionTime[128]: "ADC Acquisition Time - 128 Cycles",
    AdcAcquisitionTime[256]: "ADC Acquisition Time - 256 Cycles",
}

AdcOverSamplingRate = bidict(
    {
        1: 0,
        2: 1,
        4: 2,
        8: 3,
        16: 4,
        32: 5,
        64: 6,
        128: 7,
        256: 8,
        512: 9,
        1024: 10,
        2048: 11,
        4096: 12,
    }
)

AdcOverSamplingRateName = {
    AdcOverSamplingRate[1]: "ADC Single Sampling",
    AdcOverSamplingRate[2]: "ADC Over Sampling Rate - 2",
    AdcOverSamplingRate[4]: "ADC Over Sampling Rate - 4",
    AdcOverSamplingRate[8]: "ADC Over Sampling Rate - 8",
    AdcOverSamplingRate[16]: "ADC Over Sampling Rate - 16",
    AdcOverSamplingRate[32]: "ADC Over Sampling Rate - 32",
    AdcOverSamplingRate[64]: "ADC Over Sampling Rate - 64",
    AdcOverSamplingRate[128]: "ADC Over Sampling Rate - 128",
    AdcOverSamplingRate[256]: "ADC Over Sampling Rate - 256",
    AdcOverSamplingRate[512]: "ADC Over Sampling Rate - 512",
    AdcOverSamplingRate[1024]: "ADC Over Sampling Rate - 1024",
    AdcOverSamplingRate[2048]: "ADC Over Sampling Rate - 2048",
    AdcOverSamplingRate[4096]: "ADC Over Sampling Rate - 4096",
}

AdcReference = {
    "1V25": 25,
    "Vfs1V65": 33,
    "Vfs1V8": 36,
    "Vfs2V1": 42,
    "Vfs2V2": 44,
    "2V5": 50,
    "Vfs2V7": 54,
    "VDD": 66,
    "5V": 100,
    "6V6": 132,
}

VRefName = {
    25: "ADC Reference 1V25",
    33: "ADC Reference 1V65",
    36: "ADC Reference 1V8",
    42: "ADC Reference 2V1",
    44: "ADC Reference 2V2",
    50: "ADC Reference 2V5",
    54: "ADC Reference 2V7",
    66: "ADC Reference VDD(3V3)",
    100: "ADC Reference 5V",
    132: "ADC Reference 6V6",
}

AdcVRefValuemV = {
    AdcReference["1V25"]: 1250,
    AdcReference["Vfs1V65"]: 1650,
    AdcReference["Vfs1V8"]: 1800,
    AdcReference["Vfs2V1"]: 2100,
    AdcReference["Vfs2V2"]: 2200,
    AdcReference["2V5"]: 2500,
    AdcReference["Vfs2V7"]: 2700,
    AdcReference["VDD"]: 3300,
    AdcReference["5V"]: 5000,
    AdcReference["6V6"]: 6000,
}

DataSets = bidict(
    {
        0: 0,
        1: 1,
        3: 2,
        6: 3,
        10: 4,
        15: 5,
        20: 6,
        30: 7,
    }
)

TestCommandSignal = {
    "Line": 1,
    "Ramp": 2,
}

Prescaler = {
    "Min": 2,
    "Max": 127,
}

NodeState = bidict(
    {
        "No Change": 0,
        "Bootloader": 1,
        "Application": 2,
        "Reserved": 3,
    }
)

NetworkState = bidict(
    {
        "Failure": 0,
        "Error": 1,
        "Standby": 2,
        "GracefulDegration2": 3,
        "GracefulDegration1": 4,
        "Operating": 5,
        "Startup": 6,
        "No Change": 7,
    }
)

CalibrationFactor = {
    "Acceleration": 0,
    "Temperature": 1,
    "Voltage": 32,
}

BlueToothDeviceNr = {
    "Self": 255,
}


def byte_list_to_int(m):
    return int.from_bytes(m, "little")


def int_to_byte_list(iValue, iLength, byte_order="little"):
    return list(iValue.to_bytes(iLength, byte_order))


def calcSamplingRate(prescaler, acquisitionTime, OverSamplingRate):
    acquTime = AdcAcquisitionTime.inverse[acquisitionTime]
    samplingRate = AdcOverSamplingRate.inverse[OverSamplingRate]
    return int(
        38400000 / ((prescaler + 1) * (acquTime + 13) * samplingRate) + 0.5
    )


def payload2Hex(payload):
    payloadHex = "[{}]".format(", ".join(hex(x) for x in payload))
    return payloadHex


def AsciiStringWordBigEndian(ByteArray):
    value = 0
    for byte in range(len(ByteArray)):
        value += ByteArray[byte] << (8 * byte)
    return value


def AsciiStringWordLittleEndian(ByteArray):
    value = 0
    for byte in range(len(ByteArray)):
        value += ByteArray[byte] << (8 * (len(ByteArray) - byte - 1))
    return value


def sArray2String(Name):
    for i in range(0, len(Name)):
        Name[i] = chr(Name[i])
        i += 1
    Name = "".join(Name)
    for character in range(0, ord(" ")):
        Name = Name[0:8].replace(chr(character), "")
    for character in range(128, 0xFF):
        Name = Name[0:8].replace(chr(character), "")
    return Name


def int_to_mac_address(iAddr):
    return ":".join(
        f"{byte:02x}" for byte in int_to_byte_list(iAddr, 6, "big")
    )



import ctypes

c_uint8 = ctypes.c_uint8
c_uint32 = ctypes.c_uint32

TestConfig = {
    "Version": "v2.1.10",
    "AdcConfigAllPrescalerMax": 4,
    "ConTimeNormalMaxMs": 5000,
    "ConTimeSleep1MaxMs": 6000,
    "ConTimeSleep2MaxMs": 7000,
    "ConTimeMaximumMs": 15000,
    "DisconnectedCurrentMax": 0.8,  # in mA
    "EnergyMode1CurrentMax": 0.5,  # in mA
    "EnergyMode2CurrentMax": 0.48,  # in mA
    "EnergyModeMaxCurrentMax": 0.48,  # in mA
    "EnergyConnectedCurrentMax": 10,  # in mA
    "EnergyMeasuringCurrentMax": 17,  # in mA
    "EnergyMeasuringLedOffCurrentMax": 17,  # in mA
}

SleepTime = {
    "Min": 10000,
    "Reset1": 300000,
    "AdvertisementMin": 1000,
    "AdvertisementMax": 4000,
    "AdvertisementReset1": 2000,
    "Reset2": 3600000,
    "AdvertisementReset2": 4000,
}

SthModule = {
    "Streaming": 0,
}


class SthErrorWordFlags(ctypes.LittleEndianStructure):
    _fields_ = [
        ("bTxFail", c_uint32, 1),
        ("bAdcOverRun", c_uint32, 1),
        ("Reserved", c_uint32, 30),
    ]


class SthErrorWord(ctypes.Union):
    _fields_ = [("b", SthErrorWordFlags), ("asword", c_uint32)]


class SthStateWordFlags(ctypes.LittleEndianStructure):
    """Format for status word 0 of the STH/SHA

    bError: Error bit
        - 0: No Error
        - 1: Error

    u3NetworkState: Current network state of STH
        - 0: Failure
        - 1: Error
        - 2: Standby
        - 3: Graceful Degradation 2
        - 4: Graceful Degradation 1
        - 5: Operating
        - 6: Startup
        - 7: No Change

    Reserved: Reserved bits
    """

    _fields_ = [
        ("bError", c_uint32, 1),
        ("u3NetworkState", c_uint32, 3),
        ("Reserved", c_uint32, 28),
    ]


class SthStateWord(ctypes.Union):
    _fields_ = [("b", SthStateWordFlags), ("asword", c_uint32)]


def fVoltageBattery(x):
    if 0 < x:
        voltage = (x * 57 * 3.3) / (10 * 2**16)
    else:
        voltage = 0
    return voltage


def fAdcRawDat(x):
    return x
